"""
This module provides classes for implementing dual-functional sync/async methods in Python.

It includes descriptors and bound methods that can be used to create flexible
asynchronous interfaces, allowing methods to be called both synchronously and
asynchronously based on various conditions and configurations.
"""

# mypy: disable-error-code=valid-type
# mypy: disable-error-code=misc
import asyncio
import inspect
import typing
from cpython.object cimport PyObject
from cpython.ref cimport Py_DECREF, Py_INCREF
from libc.stdint cimport uintptr_t
from logging import getLogger

import typing_extensions

from a_sync._typing import AnyFn, AnyIterable, I, MaybeCoro, ModifierKwargs, P, T
from a_sync.a_sync import _descriptor, function
from a_sync.a_sync._kwargs cimport get_flag_name, is_sync
from a_sync.a_sync._helpers cimport _await
from a_sync.a_sync.function cimport _ASyncFunction, _ModifiedMixin
from a_sync.functools cimport update_wrapper

cdef extern from "weakrefobject.h":
    PyObject* PyWeakref_NewRef(PyObject*, PyObject*)

cdef extern from "pythoncapi_compat.h":
    int PyWeakref_GetRef(PyObject*, PyObject**)

if typing.TYPE_CHECKING:
    from a_sync import TaskMapping
    from a_sync.a_sync.abstract import ASyncABC
    
else:
    # Due to circ import issues we will populate these later
    ASyncABC, TaskMapping = None, None


# cdef asyncio
cdef object get_event_loop = asyncio.get_event_loop
cdef object iscoroutinefunction = asyncio.iscoroutinefunction
cdef object cancel_handle = asyncio.TimerHandle.cancel

# cdef inspect
cdef object isawaitable = inspect.isawaitable

# cdef typing
cdef object Any = typing.Any
cdef object Coroutine = typing.Coroutine
cdef object Generic = typing.Generic
cdef object Literal = typing.Literal
cdef object Optional = typing.Optional
cdef object Type = typing.Type
cdef object Union = typing.Union
cdef object overload = typing.overload

# cdef typing_extensions
cdef object Concatenate = typing_extensions.Concatenate
cdef object Self = typing_extensions.Self
cdef object Unpack = typing_extensions.Unpack


# cdef a_sync
cdef object ASyncDescriptor = _descriptor.ASyncDescriptor
cdef object ASyncFunctionAsyncDefault = function.ASyncFunctionAsyncDefault
cdef object ASyncFunctionSyncDefault = function.ASyncFunctionSyncDefault


cdef public double METHOD_CACHE_TTL = 3600


logger = getLogger(__name__)

cdef object _logger_is_enabled = logger.isEnabledFor
cdef object _logger_log = logger._log
cdef object DEBUG = 10

cdef inline void _logger_debug(str msg, tuple args):
    if _logger_is_enabled(DEBUG):
        _logger_log(DEBUG, msg, args)


class ASyncMethodDescriptor(ASyncDescriptor[I, P, T]):
    """
    A descriptor for managing methods that can be called both synchronously and asynchronously.

    This class provides the core functionality for binding methods to instances and determining
    the execution mode ("sync" or "async") based on various conditions, such as the instance type,
    the method's default setting, or specific flags passed during the method call.

    The descriptor is responsible for creating an appropriate bound method when accessed,
    which is the actual object that can be called in both synchronous and asynchronous contexts.
    It can create different types of bound methods (`ASyncBoundMethodSyncDefault`,
    `ASyncBoundMethodAsyncDefault`, or `ASyncBoundMethod`) based on the default mode or instance type.

    If the default mode is explicitly set to "sync" or "async", it creates `ASyncBoundMethodSyncDefault`
    or `ASyncBoundMethodAsyncDefault` respectively. If neither is set, it defaults to creating an
    `ASyncBoundMethod`. For instances of :class:`ASyncABC`, it checks the `__a_sync_instance_should_await__`
    attribute to decide the type of bound method to create.

    It also manages cache handles for bound methods and prevents setting or deleting the descriptor.

    Examples:
        >>> class MyClass:
        ...     @ASyncMethodDescriptor
        ...     async def my_method(self):
        ...         return "Hello, World!"
        ...
        >>> obj = MyClass()
        >>> await obj.my_method()
        'Hello, World!'
        >>> obj.my_method(sync=True)
        'Hello, World!'

    See Also:
        - :class:`ASyncBoundMethod`
        - :class:`ASyncFunction`
    """

    __wrapped__: AnyFn[P, T]
    """The unbound function which will be bound to an instance when :meth:`__get__` is called."""

    _initialized = False

    async def __call__(self, instance: I, *args: P.args, **kwargs: P.kwargs) -> T:
        """
        Asynchronously call the method.

        Args:
            instance: The instance the method is bound to.
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Examples:
            >>> descriptor = ASyncMethodDescriptor(my_async_function)
            >>> await descriptor(instance, arg1, arg2, kwarg1=value1)
        """
        # NOTE: This is only used by TaskMapping atm  # TODO: use it elsewhere
        _logger_debug(
            "awaiting %s for instance: %s args: %s kwargs: %s",
            (self, instance, args, kwargs),
        )
        return await self.__get__(instance, None)(*args, **kwargs)

    @overload
    def __get__(self, instance: None, owner: Type[I]) -> Self: ...
    @overload
    def __get__(self, instance: I, owner: Type[I]) -> "ASyncBoundMethod[I, P, T]": ...
    def __get__(
        _ModifiedMixin self, instance: Optional[I], owner: Type[I]
    ) -> Union[Self, "ASyncBoundMethod[I, P, T]"]:
        """
        Get the bound method or the descriptor itself.

        Args:
            instance: The instance to bind the method to, or None.
            owner: The owner class.

        Examples:
            >>> descriptor = ASyncMethodDescriptor(my_function)
            >>> bound_method = descriptor.__get__(instance, MyClass)
        """
        if instance is None:
            return self
        
        cdef str field_name = self.field_name
        cdef dict instance_dict = instance.__dict__
        
        try:
            bound = instance_dict[field_name]
        except KeyError:
            if ASyncABC is None:
                _import_ASyncABC()

            default = _ModifiedMixin.get_default(self)
            if default == "sync":
                bound = ASyncBoundMethodSyncDefault(
                    instance, self.__wrapped__, self.__is_async_def__, **self.modifiers._modifiers
                )
            elif default == "async":
                bound = ASyncBoundMethodAsyncDefault(
                    instance, self.__wrapped__, self.__is_async_def__, **self.modifiers._modifiers
                )
            elif isinstance(instance, ASyncABC):
                try:
                    if instance.__a_sync_instance_should_await__:
                        bound = ASyncBoundMethodSyncDefault(
                            instance,
                            self.__wrapped__,
                            self.__is_async_def__,
                            **self.modifiers._modifiers,
                        )
                    else:
                        bound = ASyncBoundMethodAsyncDefault(
                            instance,
                            self.__wrapped__,
                            self.__is_async_def__,
                            **self.modifiers._modifiers,
                        )
                except AttributeError:
                    bound = ASyncBoundMethod(
                        instance,
                        self.__wrapped__,
                        self.__is_async_def__,
                        **self.modifiers._modifiers,
                    )
            else:
                bound = ASyncBoundMethod(
                    instance, self.__wrapped__, self.__is_async_def__, **self.modifiers._modifiers
                )
            instance_dict[field_name] = bound
            _logger_debug("new bound method: %s", (bound,))
        _update_cache_timer(field_name, instance, bound)
        return bound

    def __set__(self, instance, value):
        """
        Prevent setting the descriptor.

        Args:
            instance: The instance.
            value: The value to set.

        Raises:
            RuntimeError: Always raised to prevent setting.

        Examples:
            >>> descriptor = ASyncMethodDescriptor(my_function)
            >>> descriptor.__set__(instance, value)
            RuntimeError: cannot set field_name, descriptor is what you get. sorry.
        """
        raise RuntimeError(
            f"cannot set {self.field_name}, {self} is what you get. sorry."
        )

    def __delete__(self, instance):
        """
        Prevent deleting the descriptor.

        Args:
            instance: The instance.

        Raises:
            RuntimeError: Always raised to prevent deletion.

        Examples:
            >>> descriptor = ASyncMethodDescriptor(my_function)
            >>> descriptor.__delete__(instance)
            RuntimeError: cannot delete field_name, you're stuck with descriptor forever. sorry.
        """
        raise RuntimeError(
            f"cannot delete {self.field_name}, you're stuck with {self} forever. sorry."
        )

    @property
    def __is_async_def__(self) -> bint:
        """
        Check if the wrapped function is a coroutine function.

        Examples:
            >>> descriptor = ASyncMethodDescriptor(my_function)
            >>> descriptor.__is_async_def__
            True
        """
        if not self._initialized:
            self._is_async_def = iscoroutinefunction(self.__wrapped__)
            self._initialized = True
        return self._is_async_def


cdef void _update_cache_timer(str field_name, instance: I, bound: "ASyncBoundMethod"):
    """
    Update the TTL for the cache handle for the instance.

    Args:
        instance: The instance to create a cache handle for.
        bound: The bound method we are caching.
    """
    # Handler for popping unused bound methods from bound method cache
    cdef object handle, loop
    if handle := bound._cache_handle:
        # update the timer handle
        handle._when = <double>handle._loop.time() + METHOD_CACHE_TTL
    else:
        # create and assign the timer handle
        loop = get_event_loop()
        # NOTE: use `instance.__dict__.pop` instead of `delattr` so we don't create a strong ref to `instance`
        bound._cache_handle = loop.call_at(<double>loop.time() + METHOD_CACHE_TTL, instance.__dict__.pop, field_name)


class ASyncMethodDescriptorSyncDefault(ASyncMethodDescriptor[I, P, T]):
    """
    A descriptor for :class:`ASyncBoundMethodSyncDefault` objects.

    This class extends :class:`ASyncMethodDescriptor` to provide a synchronous
    default behavior for method calls. It specifically creates `ASyncBoundMethodSyncDefault`
    instances, which are specialized versions of `ASyncBoundMethod` with synchronous default behavior.

    Examples:
        >>> class MyClass:
        ...     @ASyncMethodDescriptorSyncDefault
        ...     def my_method(self):
        ...         return "Hello, World!"
        ...
        >>> obj = MyClass()
        >>> obj.my_method()
        'Hello, World!'
        >>> coro = obj.my_method(sync=False)
        >>> coro
        <coroutine object MyClass.my_method at 0x7fb4f5fb49c0>
        >>> await coro
        'Hello, World!'

    See Also:
        - :class:`ASyncBoundMethodSyncDefault`
        - :class:`ASyncFunctionSyncDefault`
    """

    default = "sync"
    """The default mode for this bound method. Always set to "sync"."""

    any: ASyncFunctionSyncDefault[Concatenate[AnyIterable[I], P], bool]
    """Synchronous default version of the :meth:`~ASyncMethodDescriptor.any` method."""

    all: ASyncFunctionSyncDefault[Concatenate[AnyIterable[I], P], bool]
    """Synchronous default version of the :meth:`~ASyncMethodDescriptor.all` method."""

    min: ASyncFunctionSyncDefault[Concatenate[AnyIterable[I], P], T]
    """Synchronous default version of the :meth:`~ASyncMethodDescriptor.min` method."""

    max: ASyncFunctionSyncDefault[Concatenate[AnyIterable[I], P], T]
    """Synchronous default version of the :meth:`~ASyncMethodDescriptor.max` method."""

    sum: ASyncFunctionSyncDefault[Concatenate[AnyIterable[I], P], T]
    """Synchronous default version of the :meth:`~ASyncMethodDescriptor.sum` method."""

    @overload
    def __get__(
        self, instance: None, owner: Type[I] = None
    ) -> "ASyncMethodDescriptorSyncDefault[I, P, T]": ...
    @overload
    def __get__(
        self, instance: I, owner: Type[I] = None
    ) -> "ASyncBoundMethodSyncDefault[I, P, T]": ...
    def __get__(
        _ModifiedMixin self, instance: Optional[I], owner: Type[I] = None
    ) -> (
        "Union[ASyncMethodDescriptorSyncDefault, ASyncBoundMethodSyncDefault[I, P, T]]"
    ):
        """
        Get the bound method or the descriptor itself.

        Args:
            instance: The instance to bind the method to, or None.
            owner: The owner class.

        Examples:
            >>> descriptor = ASyncMethodDescriptorSyncDefault(my_function)
            >>> bound_method = descriptor.__get__(instance, MyClass)
        """
        if instance is None:
            return self
        
        cdef str field_name = self.field_name
        cdef dict instance_dict = instance.__dict__
        
        try:
            bound = instance_dict[field_name]
        except KeyError:
            bound = ASyncBoundMethodSyncDefault(
                instance, self.__wrapped__, self.__is_async_def__, **self.modifiers._modifiers
            )
            instance_dict[field_name] = bound
            _logger_debug("new bound method: %s", (bound,))
        _update_cache_timer(field_name, instance, bound)
        return bound


class ASyncMethodDescriptorAsyncDefault(ASyncMethodDescriptor[I, P, T]):
    """
    A descriptor for asynchronous methods with an asynchronous default.

    This class extends :class:`ASyncMethodDescriptor` to provide an asynchronous default
    behavior for method calls. It specifically creates `ASyncBoundMethodAsyncDefault`
    instances, which are specialized versions of `ASyncBoundMethod` with asynchronous default behavior.

    Examples:
        >>> class MyClass:
        ...     @ASyncMethodDescriptorAsyncDefault
        ...     async def my_method(self):
        ...         return "Hello, World!"
        ...
        >>> obj = MyClass()
        >>> coro = obj.my_method()
        >>> coro
        <coroutine object MyClass.my_method at 0x7fb4f5fb49c0>
        >>> await coro
        >>> obj.my_method(sync=True)
        'Hello, World!'

    See Also:
        - :class:`ASyncBoundMethodAsyncDefault`
        - :class:`ASyncFunctionAsyncDefault`
    """

    default = "async"
    """The default mode for this bound method. Always set to "async"."""

    any: ASyncFunctionAsyncDefault[Concatenate[AnyIterable[I], P], bool]
    """Asynchronous default version of the :meth:`~ASyncMethodDescriptor.any` method."""

    all: ASyncFunctionAsyncDefault[Concatenate[AnyIterable[I], P], bool]
    """Asynchronous default version of the :meth:`~ASyncMethodDescriptor.all` method."""

    min: ASyncFunctionAsyncDefault[Concatenate[AnyIterable[I], P], T]
    """Asynchronous default version of the :meth:`~ASyncMethodDescriptor.min` method."""

    max: ASyncFunctionAsyncDefault[Concatenate[AnyIterable[I], P], T]
    """Asynchronous default version of the :meth:`~ASyncMethodDescriptor.max` method."""

    sum: ASyncFunctionAsyncDefault[Concatenate[AnyIterable[I], P], T]
    """Asynchronous default version of the :meth:`~ASyncMethodDescriptor.sum` method."""

    @overload
    def __get__(
        self, instance: None, owner: Type[I]
    ) -> "ASyncMethodDescriptorAsyncDefault[I, P, T]": ...
    @overload
    def __get__(
        self, instance: I, owner: Type[I]
    ) -> "ASyncBoundMethodAsyncDefault[I, P, T]": ...
    def __get__(
        _ModifiedMixin self, instance: Optional[I], owner: Type[I]
    ) -> "Union[ASyncMethodDescriptorAsyncDefault, ASyncBoundMethodAsyncDefault[I, P, T]]":
        """
        Get the bound method or the descriptor itself.

        Args:
            instance: The instance to bind the method to, or None.
            owner: The owner class.

        Examples:
            >>> descriptor = ASyncMethodDescriptorAsyncDefault(my_function)
            >>> bound_method = descriptor.__get__(instance, MyClass)
        """
        if instance is None:
            return self
        
        cdef object bound
        cdef str field_name = self.field_name
        cdef dict instance_dict = instance.__dict__

        try:
            bound = instance_dict[field_name]
        except KeyError:
            bound = ASyncBoundMethodAsyncDefault(
                instance, self.__wrapped__, self.__is_async_def__, **self.modifiers._modifiers
            )
            instance_dict[field_name] = bound
            _logger_debug("new bound method: %s", (bound,))
        _update_cache_timer(field_name, instance, bound)
        return bound


cdef dict[object, bint] _is_a_sync_instance_cache = {}


cdef bint _is_a_sync_instance(object instance):
    """Checks if an instance is an ASync instance.

    Args:
        instance: The instance to check.

    Returns:
        A boolean indicating if the instance is an ASync instance.
    """
    cdef object instance_type = type(instance)
    cdef object instance_type_uid = id(instance_type)
    if instance_type_uid in _is_a_sync_instance_cache:
        return _is_a_sync_instance_cache[instance_type_uid]
    
    if ASyncABC is None:
        _import_ASyncABC()

    cdef bint is_a_sync = issubclass(instance_type, ASyncABC)
    _is_a_sync_instance_cache[instance_type_uid] = is_a_sync
    return is_a_sync


cdef bint _should_await(_ASyncFunction self, dict kwargs):
    """
    Determine if the method should be awaited.

    Args:
        kwargs: Keyword arguments passed to the method.

    Examples:
        >>> bound_method = ASyncBoundMethod(instance, my_function, True)
        >>> should_await = _should_await(bound_method, kwargs)
    """
    cdef str flag = get_flag_name(kwargs)
    if flag:
        return is_sync(flag, kwargs, pop_flag=True)  # type: ignore [arg-type]
    elif default := self.get_default():
        return default == "sync"
    elif _is_a_sync_instance(self.__self__):
        self.__self__: "ASyncABC"
        return self.__self__.__a_sync_should_await__(kwargs)
    return self._is_async_def


class ASyncBoundMethod(function.ASyncFunction[P, T], Generic[I, P, T]):
    """
    A bound method that can be called both synchronously and asynchronously.

    This class represents a method bound to an instance, which can be called
    either synchronously or asynchronously based on various conditions. It handles
    caching of bound methods and includes logic for determining whether to await
    the method call based on flags or default settings.

    Examples:
        >>> class MyClass:
        ...     def __init__(self, value):
        ...         self.value = value
        ...
        ...     @ASyncMethodDescriptor
        ...     async def my_method(self):
        ...         return self.value
        ...
        >>> obj = MyClass(42)
        >>> await obj.my_method()
        42
        >>> obj.my_method(sync=True)
        42

    See Also:
        - :class:`ASyncMethodDescriptor`
        - :class:`ASyncFunction`
    """

    # NOTE: this is created by the Descriptor

    _cache_handle: TimerHandle = None
    """An asyncio handle used to pop the bound method from `instance.__dict__` 5 minutes after its last use."""

    __weakself__: "ref[I]"
    """A weak reference to the instance the function is bound to."""

    __wrapped__: AnyFn[Concatenate[I, P], T]
    """The original unbound method that was wrapped."""

    __slots__ = "_is_async_def", "__weakself__"

    def __init__(
        _ASyncFunction self,
        instance: I,
        unbound: AnyFn[Concatenate[I, P], T],
        async_def: bool,
        **modifiers: Unpack[ModifierKwargs],
    ) -> None:
        """
        Initialize the bound method.

        Args:
            instance: The instance to bind the method to.
            unbound: The unbound function.
            async_def: Whether the original function is an async def.
            **modifiers: Additional modifiers for the function.

        Examples:
            >>> class MyClass:
            ...     def __init__(self, value):
            ...         self.value = value
            ...
            ...     @ASyncMethodDescriptor
            ...     async def my_method(self):
            ...         return self.value
            ...
            >>> obj = MyClass(42)
            >>> bound_method = ASyncBoundMethod(obj, MyClass.my_method, True)
        """
        # First we bind the method to a weak reference
        # - bind and create a strong reference to the cache handle method
        weakref_callback = self.__cancel_cache_handle
        # - we do this so the bound callback method does not get garbage collected until this ASyncBoundMethod dies
        self.__cancel_cache_handle = weakref_callback
        self.__weakself__ = <object>PyWeakref_NewRef(<PyObject*>instance, <PyObject*>weakref_callback)
        
        # Then we unwrap the coro_fn and rewrap it so overriding flag kwargs are handled automagically.
        if isinstance(unbound, _ASyncFunction):
            (<dict>modifiers).update((<_ASyncFunction>unbound).modifiers._modifiers)
            unbound = (<_ASyncFunction>unbound).__wrapped__
        # NOTE: the wrapped function was validated when the descriptor was initialized
        _ASyncFunction.__init__(self, unbound, _skip_validate=True, **<dict>modifiers)
        self._is_async_def = async_def
        """True if `self.__wrapped__` is a coroutine function, False otherwise."""
        update_wrapper(self, unbound)

    def __repr__(self) -> str:
        """
        Return a string representation of the bound method.

        Examples:
            >>> bound_method = ASyncBoundMethod(instance, my_function, True)
            >>> repr(bound_method)
            '<ASyncBoundMethod for function module.ClassName.method_name bound to instance>'
        """
        cdef object instance_type
        try:
            instance_type = type(self.__self__)
            return "<{} for function {}.{}.{} bound to {}>".format(
                self.__class__.__name__,
                instance_type.__module__,
                instance_type.__name__,
                self.__name__,
                self.__self__
            )
        except ReferenceError:
            return "<{} for function COLLECTED.COLLECTED.{} bound to {}>".format(
                self.__class__.__name__,
                self.__name__,
                self.__weakself__
            )

    @overload
    def __call__(self, *args: P.args, sync: Literal[True], **kwargs: P.kwargs) -> T: ...
    @overload
    def __call__(
        self, *args: P.args, sync: Literal[False], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]: ...
    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[False], **kwargs: P.kwargs
    ) -> T: ...
    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[True], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]: ...
    @overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> MaybeCoro[T]: ...
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> MaybeCoro[T]:
        """
        Call the bound method.

        This method handles both synchronous and asynchronous calls based on
        the provided flags and the method's configuration.

        Args:
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Examples:
            >>> bound_method = ASyncBoundMethod(instance, my_function, True)
            >>> await bound_method(arg1, arg2, kwarg1=value1)
            >>> bound_method(arg1, arg2, kwarg1=value1, sync=True)
        """
        cdef object retval, coro
        cdef bint debug_logs
        if debug_logs := _logger_is_enabled(DEBUG):
            _logger_log(DEBUG, "calling %s with args: %s kwargs: %s", (self, args, kwargs))
        # This could either be a coroutine or a return value from an awaited coroutine,
        #   depending on if an overriding flag kwarg was passed into the function call.
        retval = coro = _ASyncFunction.__call__(self, self.__self__, *args, **kwargs)
        if not isawaitable(retval):
            # The coroutine was already awaited due to the use of an overriding flag kwarg.
            # We can return the value.
            pass
        elif _should_await(self, kwargs):
            # The awaitable was not awaited, so now we need to check the flag as defined on 'self' and await if appropriate.
            if debug_logs:
                _logger_log(
                    DEBUG, "awaiting %s for %s args: %s kwargs: %s", (coro, self, args, kwargs)
                )
            retval = _await(coro)
        if debug_logs:
            _logger_log(
                DEBUG, "returning %s for %s args: %s kwargs: %s", (retval, self, args, kwargs)
            )
        return retval  # type: ignore [call-overload, return-value]

    @property
    def __self__(self) -> I:
        """
        Get the instance the method is bound to.

        Raises:
            ReferenceError: If the instance has been garbage collected.

        Examples:
            >>> bound_method = ASyncBoundMethod(instance, my_function, True)
            >>> bound_method.__self__
            <MyClass instance>
        """
        cdef PyObject *self_ptr
        if PyWeakref_GetRef(<PyObject*>self.__weakself__, &self_ptr) == 1:
            # 1 is success
            return <object>self_ptr
        raise ReferenceError(self)

    def map(
        self,
        *iterables: AnyIterable[I],
        concurrency: Optional[int] = None,
        task_name: str = "",
        **kwargs: P.kwargs,
    ) -> "TaskMapping[I, T]":
        """
        Create a TaskMapping for this method.

        Args:
            *iterables: Iterables to map over.
            concurrency: Optional concurrency limit.
            task_name: Optional name for the task.
            **kwargs: Additional keyword arguments.

        Returns:
            A TaskMapping instance for this method.

        Examples:
            >>> bound_method = ASyncBoundMethod(instance, my_function, True)
            >>> task_mapping = bound_method.map(iterable1, iterable2, concurrency=5)
            TODO briefly include how someone would then use task_mapping
        """
        if TaskMapping is None:
            _import_TaskMapping()

        return TaskMapping(
            self, *iterables, concurrency=concurrency, name=task_name, **kwargs
        )

    async def any(
        self,
        *iterables: AnyIterable[I],
        concurrency: Optional[int] = None,
        task_name: str = "",
        **kwargs: P.kwargs,
    ) -> bool:
        """
        Check if any of the results are truthy.

        Args:
            *iterables: Iterables to map over.
            concurrency: Optional concurrency limit.
            task_name: Optional name for the task.
            **kwargs: Additional keyword arguments.

        Examples:
            >>> bound_method = ASyncBoundMethod(instance, my_function, True)
            >>> result = await bound_method.any(iterable1, iterable2)
        """
        return await self.map(
            *iterables, concurrency=concurrency, task_name=task_name, **kwargs
        ).any(pop=True, sync=False)

    async def all(
        self,
        *iterables: AnyIterable[I],
        concurrency: Optional[int] = None,
        task_name: str = "",
        **kwargs: P.kwargs,
    ) -> bool:
        """
        Check if all of the results are truthy.

        Args:
            *iterables: Iterables to map over.
            concurrency: Optional concurrency limit.
            task_name: Optional name for the task.
            **kwargs: Additional keyword arguments.

        Examples:
            >>> bound_method = ASyncBoundMethod(instance, my_function, True)
            >>> result = await bound_method.all(iterable1, iterable2)
        """
        return await self.map(
            *iterables, concurrency=concurrency, task_name=task_name, **kwargs
        ).all(pop=True, sync=False)

    async def min(
        self,
        *iterables: AnyIterable[I],
        concurrency: Optional[int] = None,
        task_name: str = "",
        **kwargs: P.kwargs,
    ) -> T:
        """
        Find the minimum result.

        Args:
            *iterables: Iterables to map over.
            concurrency: Optional concurrency limit.
            task_name: Optional name for the task.
            **kwargs: Additional keyword arguments.

        Examples:
            >>> bound_method = ASyncBoundMethod(instance, my_function, True)
            >>> result = await bound_method.min(iterable1, iterable2)
        """
        return await self.map(
            *iterables, concurrency=concurrency, task_name=task_name, **kwargs
        ).min(pop=True, sync=False)

    async def max(
        self,
        *iterables: AnyIterable[I],
        concurrency: Optional[int] = None,
        task_name: str = "",
        **kwargs: P.kwargs,
    ) -> T:
        """
        Find the maximum result.

        Args:
            *iterables: Iterables to map over.
            concurrency: Optional concurrency limit.
            task_name: Optional name for the task.
            **kwargs: Additional keyword arguments.

        Examples:
            >>> bound_method = ASyncBoundMethod(instance, my_function, True)
            >>> result = await bound_method.max(iterable1, iterable2)
        """
        return await self.map(
            *iterables, concurrency=concurrency, task_name=task_name, **kwargs
        ).max(pop=True, sync=False)

    async def sum(
        self,
        *iterables: AnyIterable[I],
        concurrency: Optional[int] = None,
        task_name: str = "",
        **kwargs: P.kwargs,
    ) -> T:
        """
        Calculate the sum of the results.

        Args:
            *iterables: Iterables to map over.
            concurrency: Optional concurrency limit.
            task_name: Optional name for the task.
            **kwargs: Additional keyword arguments.

        Examples:
            >>> bound_method = ASyncBoundMethod(instance, my_function, True)
            >>> result = await bound_method.sum(iterable1, iterable2)
        """
        return await self.map(
            *iterables, concurrency=concurrency, task_name=task_name, **kwargs
        ).sum(pop=True, sync=False)

    def __cancel_cache_handle(self, instance: I) -> None:
        """
        Cancel the cache handle.

        Args:
            instance: The instance associated with the cache handle.

        Examples:
            >>> bound_method = ASyncBoundMethod(instance, my_function, True)
            >>> bound_method.__cancel_cache_handle(instance)
        """
        handle = self._cache_handle
        if handle is not None:
            cancel_handle(handle)


class ASyncBoundMethodSyncDefault(ASyncBoundMethod[I, P, T]):
    """
    A bound method with synchronous default behavior.

    This class is a specialized version of :class:`ASyncBoundMethod` that defaults to synchronous execution.
    It overrides the `__call__` method to enforce synchronous default behavior.

    Examples:
        >>> class MyClass:
        ...     def __init__(self, value):
        ...         self.value = value
        ...
        ...     @ASyncMethodDescriptorSyncDefault
        ...     def my_method(self):
        ...         return self.value
        ...
        >>> obj = MyClass(42)
        >>> obj.my_method()
        42
        >>> await obj.my_method(sync=False)
        42

    See Also:
        - :class:`ASyncBoundMethod`
        - :class:`ASyncMethodDescriptorSyncDefault`
    """

    @overload
    def __call__(self, *args: P.args, sync: Literal[True], **kwargs: P.kwargs) -> T: ...
    @overload
    def __call__(
        self, *args: P.args, sync: Literal[False], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]: ...
    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[False], **kwargs: P.kwargs
    ) -> T: ...
    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[True], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]: ...
    @overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T: ...

    __call__ = ASyncBoundMethod.__call__
    """
    Call the bound method with synchronous default behavior.

    Args:
        *args: Positional arguments.
        **kwargs: Keyword arguments.

    Examples:
        >>> bound_method = ASyncBoundMethodSyncDefault(instance, my_function, True)
        >>> bound_method(arg1, arg2, kwarg1=value1)
    """


class ASyncBoundMethodAsyncDefault(ASyncBoundMethod[I, P, T]):
    """
    A bound method with asynchronous default behavior.

    This class is a specialized version of :class:`ASyncBoundMethod` that defaults to asynchronous execution.
    It overrides the `__call__` method to enforce asynchronous default behavior.

    Examples:
        >>> class MyClass:
        ...     def __init__(self, value):
        ...         self.value = value
        ...
        ...     @ASyncMethodDescriptorAsyncDefault
        ...     async def my_method(self):
        ...         return self.value
        ...
        >>> obj = MyClass(42)
        >>> await obj.my_method()
        42
        >>> obj.my_method(sync=True)
        42

    See Also:
        - :class:`ASyncBoundMethod`
        - :class:`ASyncMethodDescriptorAsyncDefault`
    """

    @overload
    def __call__(self, *args: P.args, sync: Literal[True], **kwargs: P.kwargs) -> T: ...
    @overload
    def __call__(
        self, *args: P.args, sync: Literal[False], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]: ...
    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[False], **kwargs: P.kwargs
    ) -> T: ...
    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[True], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]: ...
    @overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Coroutine[Any, Any, T]: ...

    __call__ = ASyncBoundMethod.__call__
    """
    Call the bound method with asynchronous default behavior.

    Args:
        *args: Positional arguments.
        **kwargs: Keyword arguments.

    Examples:
        >>> bound_method = ASyncBoundMethodAsyncDefault(instance, my_function, True)
        >>> await bound_method(arg1, arg2, kwarg1=value1)
    """


cdef inline void _import_ASyncABC():
    global ASyncABC
    from a_sync.a_sync.abstract import ASyncABC


cdef inline void _import_TaskMapping():
    global TaskMapping
    from a_sync import TaskMapping


del asyncio, inspect, typing, typing_extensions
del _descriptor, function




"""
This module extends :func:`asyncio.create_task` to support any :class:`Awaitable`,
manage task lifecycle, and enhance error handling.
"""

import asyncio
import asyncio.tasks as aiotasks
import logging
import typing

from cpython.object cimport PyObject
from cpython.unicode cimport PyUnicode_CompareWithASCIIString

from a_sync import _smart, exceptions
from a_sync._typing import T


# cdef asyncio
cdef object get_running_loop = asyncio.get_running_loop
cdef object iscoroutine = asyncio.iscoroutine
cdef object Future = asyncio.Future
cdef object InvalidStateError = asyncio.InvalidStateError
cdef object Task = asyncio.Task
cdef object _GatheringFuture = aiotasks._GatheringFuture
del asyncio, aiotasks

# cdef logging
cdef public object logger = logging.getLogger(__name__)
del logging

# cdef typing
cdef object Awaitable = typing.Awaitable
cdef object Optional = typing.Optional
cdef object Tuple = typing.Tuple
cdef object Set = typing.Set
del typing

# cdef _smart
cdef object SmartTask = _smart.SmartTask
cdef object smart_task_factory = _smart.smart_task_factory
del _smart

# cdef exceptions
cdef object PersistedTaskException = exceptions.PersistedTaskException
del exceptions


def create_task(
    coro: Awaitable[T],
    *,
    name: str = "",
    skip_gc_until_done: bint = False,
    log_destroy_pending: bint = True,
) -> "Task[T]":
    """
    Extends :func:`asyncio.create_task` to support any :class:`Awaitable`, manage task lifecycle, and enhance error handling.

    This function accepts any :class:`Awaitable`, ensuring broader compatibility. If the Awaitable is not a coroutine,
    it is awaited directly using a private helper function `__await`, which can handle non-coroutine Awaitable objects.

    Note:
        The `__await` function is designed to handle non-coroutine Awaitables by awaiting them directly.

    Args:
        coro: An :class:`Awaitable` object from which to create the task.
        name: Optional name for the task, aiding in debugging.
        skip_gc_until_done: If True, the task is kept alive until it completes, preventing garbage collection.
            Exceptions are wrapped in :class:`PersistedTaskException` for special handling within the
            `__persisted_task_exc_wrap` function.
        log_destroy_pending: If False, asyncio's default error log when a pending task is destroyed is suppressed.

    Examples:
        Create a simple task with a coroutine:

        >>> async def my_coroutine():
        ...     return "Hello, World!"
        >>> task = create_task(my_coroutine())

        Create a task with a non-coroutine Awaitable:

        >>> from concurrent.futures import Future
        >>> future = Future()
        >>> task = create_task(future)

    See Also:
        - :func:`asyncio.create_task`
        - :class:`Task`
    """
    return ccreate_task(coro, name, skip_gc_until_done, log_destroy_pending)

cdef object ccreate_task_simple(object coro):
    return ccreate_task(coro, "", False, True)
    
cdef object ccreate_task(object coro, str name, bint skip_gc_until_done, bint log_destroy_pending):
    cdef object loop = get_running_loop()
    cdef object task_factory = loop._task_factory
    cdef object task, persisted
    
    if not iscoroutine(coro):
        coro = __await(coro)
    
    if task_factory is None:
        task = Task(coro, loop=loop, name=name)
        if task._source_traceback:
            del task._source_traceback[-1]
    elif task_factory is smart_task_factory:
        task = SmartTask(coro, loop=loop, name=name)
        if task._source_traceback:
            del task._source_traceback[-1]
    else:
        task = task_factory(loop, coro)
        if name:
            __set_task_name(task, name)

    if skip_gc_until_done:
        persisted = __persisted_task_exc_wrap(task)
            
        if task_factory is None:
            persisted = Task(persisted, loop=loop, name=name)
            if persisted._source_traceback:
                del persisted._source_traceback[-1]
        elif task_factory is smart_task_factory:
            persisted = SmartTask(persisted, loop=loop, name=name)
            if persisted._source_traceback:
                del persisted._source_traceback[-1]
        else:
            persisted = task_factory(loop, persisted)
            if name:
                __set_task_name(persisted, name)
            
        _persisted_tasks.add(persisted)

    if log_destroy_pending is False:
        task._log_destroy_pending = False

    if _persisted_tasks:
        __prune_persisted_tasks()

    return task


cdef inline void __set_task_name(object task, str name):
    if set_name := getattr(task, "set_name", None):
         set_name(name)


cdef public set[object] _persisted_tasks = set()

cdef object __await

async def __await(awaitable: Awaitable[T]) -> T:
    """Wait for the completion of a non-coroutine Awaitable.

    Args:
        awaitable: The :class:`Awaitable` object to wait for.

    Raises:
        RuntimeError: If a RuntimeError occurs during the await, it is raised with additional context.

    Examples:
        Await a simple coroutine:

        >>> async def my_coroutine():
        ...     return "Hello, World!"
        >>> result = await __await(my_coroutine())

    See Also:
        - :class:`Awaitable`
    """
    try:
        return await awaitable
    except RuntimeError as e:
        args = [e, awaitable]
        if isinstance(awaitable, _GatheringFuture):
            args.append(awaitable._children)
        raise RuntimeError(*args) from None

   
cdef object __log_exception = logger.exception


cdef void __prune_persisted_tasks():
    """Remove completed tasks from the set of persisted tasks.

    This function checks each task in the persisted tasks set. If a task is done and has an exception,
    it logs the exception and raises it if it's not a :class:`PersistedTaskException`. It also logs the traceback
    manually since the usual handler will not run after retrieving the exception.

    See Also:
        - :class:`PersistedTaskException`
    """
    cdef object task
    cdef dict context
    cdef list done = list(filter(_is_done, _persisted_tasks))
    if not done:
        return
    _persisted_tasks.difference_update(done)
    for task in done:
        exc = _get_exception(task)
        if exc is None:
            continue
        # force exceptions related to this lib to bubble up
        if not isinstance(exc, PersistedTaskException):
            __log_exception(exc)
            raise exc
        # we have to manually log the traceback that asyncio would usually log
        # since we already got the exception from the task and the usual handler will now not run
        context = {
            "message": f"{task.__class__.__name__} exception was never retrieved",
            "exception": exc,
            "future": task,
        }
        if task._source_traceback:
            context["source_traceback"] = task._source_traceback
        task._loop.call_exception_handler(context)


cdef inline bint _is_done(fut: Future):
    return PyUnicode_CompareWithASCIIString(fut._state, b"PENDING") != 0


cdef object _get_exception(fut: Future):
    """Return the exception that was set on this future.

    The exception (or None if no exception was set) is returned only if
    the future is done.  If the future has been cancelled, raises
    CancelledError.  If the future isn't done yet, raises
    InvalidStateError.
    """
    cdef str state = fut._state
    if PyUnicode_CompareWithASCIIString(state, b"FINISHED") == 0:
        fut._Future__log_traceback = False
        return fut._exception
    if PyUnicode_CompareWithASCIIString(state, b"CANCELLED") == 0:
        raise fut._make_cancelled_error()
    raise InvalidStateError('Exception is not set.')


cdef inline bint _exc_exists(tup: Tuple[Future, Optional[Exception]]):
    return tup[1] is not None


cdef object __persisted_task_exc_wrap

async def __persisted_task_exc_wrap(task: "Task[T]") -> T:
    """
    Wrap a task to handle its exception in a specialized manner.

    Args:
        task: The :class:`Task` to wrap.

    Raises:
        PersistedTaskException: Wraps any exception raised by the task for special handling.

    See Also:
        - :class:`PersistedTaskException`
    """
    try:
        return await task
    except Exception as e:
        raise PersistedTaskException(e, task) from e


__all__ = ["create_task"]


# For testing purposes only

def _get_persisted_tasks() -> Set[Task]:
    # we can't import this directly to the .py test file
    return _persisted_tasks



# cython: profile=False
# cython: linetrace=False
import asyncio
import inspect
import functools
import time
from logging import DEBUG, Logger, getLogger
from typing import AsyncIterator, Callable, NoReturn, TypeVar

from typing_extensions import Concatenate, ParamSpec

from a_sync.a_sync.base import ASyncGenericBase
from a_sync.iter cimport _ASyncGeneratorFunction


__P = ParamSpec("__P")
__T = TypeVar("__T")
__B = TypeVar("__B", bound=ASyncGenericBase)


cdef object _FIVE_MINUTES = 300

cdef object create_task = asyncio.create_task
cdef object sleep = asyncio.sleep

cdef object isasyncgenfunction = inspect.isasyncgenfunction

cdef object wraps = functools.wraps

cdef object gettime = time.time

cdef object logger = getLogger("a_sync.debugging")


def stuck_coro_debugger(fn, logger=logger, interval=_FIVE_MINUTES):
    __logger_is_enabled_for = logger.isEnabledFor

    if isasyncgenfunction(fn) or isinstance(fn, _ASyncGeneratorFunction):

        @wraps(fn)
        async def stuck_async_gen_wrap(*args: __P.args, **kwargs: __P.kwargs) -> AsyncIterator[__T]:
            aiterator = fn(*args, **kwargs)

            if not __logger_is_enabled_for(DEBUG):
                async for thing in aiterator:
                    yield thing
                return

            task = create_task(
                coro=_stuck_debug_task(logger, interval, fn, args, kwargs),
                name="_stuck_debug_task",
            )
            try:
                async for thing in aiterator:
                    yield thing
            finally:
                task.cancel()

        return stuck_async_gen_wrap
    else:

        @wraps(fn)
        async def stuck_coro_wrap(*args: __P.args, **kwargs: __P.kwargs) -> __T:
            if not __logger_is_enabled_for(DEBUG):
                return await fn(*args, **kwargs)

            task = create_task(
                coro=_stuck_debug_task(logger, interval, fn, args, kwargs),
                name="_stuck_debug_task",
            )
            try:
                retval = await fn(*args, **kwargs)
            finally:
                task.cancel()
            return retval

        return stuck_coro_wrap


async def _stuck_debug_task(
    logger: Logger, interval: int, fn: Callable[__P, __T], *args: __P.args, **kwargs: __P.kwargs
) -> NoReturn:
    # sleep early so fast-running coros can exit early
    await sleep(interval)

    start = gettime() - interval
    cdef str module = fn.__module__
    cdef str name = fn.__name__
    cdef tuple formatted_args = tuple(map(str, args))
    cdef dict formatted_kwargs = dict(zip(kwargs.keys(), map(str, kwargs.values())))
    log = logger._log
    while True:
        log(
            DEBUG,
            "%s.%s still executing after %sm with args %s kwargs %s",
            (
                module,
                name,
                round((gettime() - start) / 60, 2),
                formatted_args,
                formatted_kwargs,
            ),
        )
        await sleep(interval)


__all__ = ["stuck_coro_debugger"]



from itertools import islice
from typing import Any, Iterable

from cpython.dict cimport PyDict_Size
from cpython.list cimport PyList_GET_SIZE
from cpython.object cimport Py_TYPE, PyObject_Repr
from cpython.tuple cimport PyTuple_GET_SIZE

cdef extern from "Python.h":
    ctypedef struct PyTypeObject:
        pass


L, S, D = [], set(), {}
cdef PyTypeObject *List_ptr = Py_TYPE(L)
cdef PyTypeObject *Tuple_ptr = Py_TYPE(())
cdef PyTypeObject *Dict_ptr = Py_TYPE(D)
cdef PyTypeObject *DictKeys_ptr = Py_TYPE(D.keys())
cdef PyTypeObject *DictValues_ptr = Py_TYPE(D.values())
cdef PyTypeObject *DictItems_ptr = Py_TYPE(D.items())
cdef PyTypeObject *Set_ptr = Py_TYPE(S)
del L, S, D
    

def repr_trunc(iterable: Iterable[Any]) -> str:
    """Returns a truncated `repr` for `iterable`, limited to the first 5 items."""
    cdef PyTypeObject *itype = Py_TYPE(iterable)
    if itype == List_ptr:
        return _join_first_5_reprs_list(iterable)
    elif itype == Tuple_ptr:
        return _join_first_5_reprs_tuple(iterable)
    elif itype == Dict_ptr:
        if PyDict_Size(iterable) <= 5:
            return "{" + ", ".join(
                f"{PyObject_Repr(k)}: {repr(iterable[k])}"
                for k in iterable
            ) + "}"
        else:
            return "{" + ", ".join(
                f"{PyObject_Repr(k)}: {repr(iterable[k])}"
                for k in islice(iterable, 5)
            ) + ", ...}"
    elif itype == DictKeys_ptr:
        return f"dict_keys([{_join_first_5_reprs_generic(iterable)}])"
    elif itype == DictValues_ptr:
        return f"dict_values([{_join_first_5_reprs_generic(iterable)}])"
    elif itype == DictItems_ptr:
        return f"dict_items([{_join_first_5_reprs_generic(iterable)}])"
    elif itype == Set_ptr:
        return "{" + _join_first_5_reprs_generic(iterable) + "}"
    else:
        return PyObject_Repr(iterable)


cdef inline str _join_first_5_reprs_list(list[object] lst):
    if PyList_GET_SIZE(lst) <= 5:
        return f"[{', '.join(PyObject_Repr(obj) for obj in lst)}]"
    else:
        return f"[{', '.join(PyObject_Repr(obj) for obj in lst[:5])}, ...]"


cdef inline str _join_first_5_reprs_tuple(tuple[object, ...] tup):
    if PyTuple_GET_SIZE(tup) <= 5:
        return f"({', '.join(PyObject_Repr(obj) for obj in tup)})"
    else:
        return f"({', '.join(PyObject_Repr(obj) for obj in tup[:5])}, ...)"


cdef str _join_first_5_reprs_generic(iterable: Iterable[Any]):
    if len(iterable) <= 5:
        return ", ".join(PyObject_Repr(obj) for obj in iterable)
    else:
        return f"{', '.join(PyObject_Repr(obj) for obj in islice(iterable, 5))}, ..."



import pytest

from a_sync import property


class PropertyTester:
    def __init__(self, i: int):
        self.i = i

    @property
    async def do_stuff(self):
        return self.i


testers = [PropertyTester(i) for i in range(5)]


def test_property_descriptor_any_sync():
    assert PropertyTester.do_stuff.any(testers, sync=True) == True


@pytest.mark.asyncio_cooperative
async def test_property_descriptor_any_async():
    assert await PropertyTester.do_stuff.any(testers, sync=False) == True


def test_property_descriptor_all_sync():
    assert PropertyTester.do_stuff.all(testers, sync=True) == False


@pytest.mark.asyncio_cooperative
async def test_property_descriptor_all_async():
    assert await PropertyTester.do_stuff.all(testers, sync=False) == False


def test_property_descriptor_min_sync():
    assert PropertyTester.do_stuff.min(testers, sync=True) == 0


@pytest.mark.asyncio_cooperative
async def test_property_descriptor_min_async():
    assert await PropertyTester.do_stuff.min(testers, sync=False) == 0


def test_property_descriptor_max_sync():
    assert PropertyTester.do_stuff.max(testers, sync=True) == 4


@pytest.mark.asyncio_cooperative
async def test_property_descriptor_max_async():
    assert await PropertyTester.do_stuff.max(testers, sync=False) == 4


def test_property_descriptor_sum_sync():
    assert PropertyTester.do_stuff.sum(testers, sync=True) == 10


@pytest.mark.asyncio_cooperative
async def test_property_descriptor_sum_async():
    assert await PropertyTester.do_stuff.sum(testers, sync=False) == 10



from typing import Optional


cdef object _NOT_FOUND = object()

cdef class cached_property_unsafe:
    """A non-threadsafe implementation of functools.cached_property, intended for use in asyncio applications"""

    def __cinit__(self, object func):
        self.func = func
        self.attrname = None
        self._doc = func.__doc__
    
    @property
    def __doc__(self) -> Optional[str]:
        return self._doc

    def __set_name__(self, owner, str name):
        cdef str existing_name = self.attrname
        if existing_name is None:
            self.attrname = name
        elif name != existing_name:
            raise TypeError(
                "Cannot assign the same cached_property to two different names "
                f"({existing_name!r} and {name!r})."
            )

    def __get__(self, instance, owner):
        cdef str attrname
        cdef dict cache
        if instance is None:
            return self
        attrname = self.attrname
        if attrname is None:
            raise TypeError(
                "Cannot use cached_property instance without calling __set_name__ on it.")
        try:
            cache = instance.__dict__
        except AttributeError:  # not all objects have __dict__ (e.g. class defines slots)
            msg = (
                f"No '__dict__' attribute on {type(instance).__name__!r} "
                f"instance to cache {attrname!r} property."
            )
            raise TypeError(msg) from None
        val = cache.get(attrname, _NOT_FOUND)
        if val is _NOT_FOUND:
            val = self.func(instance)
            try:
                cache[attrname] = val
            except TypeError:
                msg = (
                    f"The '__dict__' attribute on {type(instance).__name__!r} instance "
                    f"does not support item assignment for caching {attrname!r} property."
                )
                raise TypeError(msg) from None
        return val


################################################################################
### update_wrapper() and wraps() decorator
################################################################################

# update_wrapper() and wraps() are tools to help write
# wrapper functions that can handle naive introspection

cpdef update_wrapper(wrapper, wrapped):
    """Update a wrapper function to look like the wrapped function

    wrapper is the function to be updated
    wrapped is the original function
    assigned is a tuple naming the attributes assigned directly
    from the wrapped function to the wrapper function (defaults to
    functools.WRAPPER_ASSIGNMENTS)
    updated is a tuple naming the attributes of the wrapper that
    are updated with the corresponding attribute from the wrapped
    function (defaults to functools.WRAPPER_UPDATES)

    `assigned` and `updated` args from the functools implementation have
    been disabled for faster tight loops. Use the implementation in the
    `functools` builtin module if you need to use their functionality.
    """
    try:
        value = getattr(wrapped, '__module__')
    except AttributeError:
        pass
    else:
        setattr(wrapper, '__module__', value)

    try:
        value = getattr(wrapped, '__name__')
    except AttributeError:
        pass
    else:
        setattr(wrapper, '__name__', value)

    try:
        value = getattr(wrapped, '__qualname__')
    except AttributeError:
        pass
    else:
        setattr(wrapper, '__qualname__', value)

    try:
        value = getattr(wrapped, '__doc__')
    except AttributeError:
        pass
    else:
        setattr(wrapper, '__doc__', value)

    try:
        value = getattr(wrapped, '__annotations__')
    except AttributeError:
        pass
    else:
        setattr(wrapper, '__annotations__', value)

    getattr(wrapper, '__dict__').update(getattr(wrapped, '__dict__', {}))
    # Issue #17482: set __wrapped__ last so we don't inadvertently copy it
    # from the wrapped function when updating __dict__
    wrapper.__wrapped__ = wrapped
    # Return the wrapper so this can be used as a decorator via partial()
    return wrapper


cdef wraps(wrapped):
    """Decorator factory to apply update_wrapper() to a wrapper function

    Returns a decorator that invokes update_wrapper() with the decorated
    function as the wrapper argument and the arguments to wraps() as the
    remaining arguments. Default arguments are as for update_wrapper().
    This is a convenience function to simplify applying partial() to
    update_wrapper().

    `assigned` and `updated` args from the functools implementation have
    been disabled for faster tight loops. Use the implementation in the
    `functools` builtin module if you need to use their functionality.

    """
    return lambda wrapper: update_wrapper(wrapper, wrapped)



import asyncio
import inspect
import sys
import typing
from logging import getLogger
from libc.stdint cimport uintptr_t

import async_lru
import async_property
from typing_extensions import Unpack

from a_sync._typing import AnyFn, AnyIterable, CoroFn, DefaultMode, MaybeCoro, ModifierKwargs, P, SyncFn, T
from a_sync.a_sync._kwargs cimport get_flag_name, is_sync
from a_sync.a_sync._helpers cimport _asyncify, _await
from a_sync.a_sync.flags cimport VIABLE_FLAGS
from a_sync.a_sync.modifiers cimport ModifierManager
from a_sync.functools cimport update_wrapper, wraps

if typing.TYPE_CHECKING:
    from a_sync import TaskMapping
    from a_sync.a_sync.method import (
        ASyncBoundMethod,
        ASyncBoundMethodAsyncDefault,
        ASyncBoundMethodSyncDefault,
    )

else:
    # due to circ import issues we will populate this later
    TaskMapping = None


ctypedef object object_id


# cdef asyncio
cdef object iscoroutinefunction = asyncio.iscoroutinefunction
del asyncio

# cdef inspect
cdef object getargspec
if sys.version_info < (3, 11):
    getargspec = inspect.getargspec
else:
    # getargspec was deprecated in python 3.11
    getargspec = inspect.getfullargspec
cdef object isasyncgenfunction = inspect.isasyncgenfunction
cdef object isgeneratorfunction = inspect.isgeneratorfunction
del inspect

# cdef logging
cdef public object logger = getLogger(__name__)
cdef object _logger_debug = logger.debug

# cdef typing
cdef object TYPE_CHECKING = typing.TYPE_CHECKING
cdef object Any = typing.Any
cdef object Callable = typing.Callable
cdef object Coroutine = typing.Coroutine
cdef object Generic = typing.Generic
cdef object Literal = typing.Literal
cdef object Optional = typing.Optional
cdef object overload = typing.overload


# cdef async_lru
cdef object _LRUCacheWrapper = async_lru._LRUCacheWrapper

# cdef async_property
cdef object AsyncPropertyDescriptor = async_property.base.AsyncPropertyDescriptor
cdef object AsyncCachedPropertyDescriptor = async_property.cached.AsyncCachedPropertyDescriptor


cdef class _ModifiedMixin:
    """
    A mixin class for internal use that provides functionality for applying modifiers to functions.

    This class is used as a base for :class:`~ASyncFunction` and its variants, such as
    `ASyncFunctionAsyncDefault` and `ASyncFunctionSyncDefault`, to handle the application
    of async and sync modifiers to functions. Modifiers can alter the behavior of functions,
    such as converting sync functions to async, applying caching, or rate limiting.

    See Also:
        - :class:`~ASyncFunction`
        - :class:`~ModifierManager`
    """

    @property
    def default(self) -> DefaultMode:
        """
        Gets the default execution mode (sync, async, or None) for the function.

        Returns:
            The default execution mode.

        See Also:
            - :attr:`ModifierManager.default`
        """
        return self.get_default()

    cdef inline object _asyncify(self, func: SyncFn[P, T]):
        """
        Converts a synchronous function to an asynchronous one and applies async modifiers.

        Args:
            func: The synchronous function to be converted.

        Returns:
            The asynchronous version of the function with applied modifiers.

        See Also:
            - :meth:`ModifierManager.apply_async_modifiers`
        """
        return self.modifiers.apply_async_modifiers(_asyncify(func, self.modifiers.executor))
    
    cdef inline object get_await(self):
        """
        Applies sync modifiers to the _helpers._await function and caches it.

        Returns:
            The modified _await function.

        See Also:
            - :meth:`ModifierManager.apply_sync_modifiers`
        """
        awaiter = self.__await
        if awaiter is None:
            awaiter = self.modifiers.apply_sync_modifiers(_await)
            self.__await = awaiter
        return awaiter
    
    cdef inline str get_default(self):
        cdef str default = self.__default
        if default is None:
            default = self.modifiers.get_default()
            self.__default = default
        return default



cdef void _validate_wrapped_fn(fn: Callable):
    """Ensures 'fn' is an appropriate function for wrapping with a_sync.

    Args:
        fn: The function to validate.

    Raises:
        TypeError: If the input is not callable.
        RuntimeError: If the function has arguments with names that conflict with viable flags.

    See Also:
        - :func:`_check_not_genfunc`
    """
    typ = type(fn)
    if typ is _function_type:
        _check_not_genfunc_cached(fn)
        _validate_argspec_cached(fn)
        return
    if issubclass(typ, (AsyncPropertyDescriptor, AsyncCachedPropertyDescriptor)):
        return  # These are always valid
    elif issubclass(typ, _LRUCacheWrapper):
        fn = fn.__wrapped__
        if type(fn) is _function_type:
            _check_not_genfunc_cached(fn)
            _validate_argspec_cached(fn)
            return
    elif not callable(fn):
        raise TypeError(f"Input is not callable. Unable to decorate {fn}")
    _check_not_genfunc(fn)
    _validate_argspec(fn)

cdef object _function_type = type(getLogger)

cdef set[object_id] _argspec_validated = set()

cdef inline void _validate_argspec_cached(fn: Callable):
    cdef object_id fid = id(fn)
    if fid not in _argspec_validated:
        _validate_argspec(fn)
        _argspec_validated.add(fid)

cdef inline void _validate_argspec(fn: Callable):
    cdef tuple[str, ...] fn_args
    cdef object fn_code
    
    try:
        fn_code = fn.__code__  # May fail for built-ins or special callables
        fn_args = fn_code.co_varnames[:fn_code.co_argcount]
    except:
        try:
            argspec = getargspec(fn)
        except TypeError:
            warn = logger.warning
            warn(f"inspect.{getargspec.__name__} does not support {fn}")
            warn("we will allow you to proceed but cannot guarantee things will work")
            warn("hopefully you know what you're doing...")
            return
        else:
            fn_args = argspec[0]
    
    for flag in fn_args:
        if flag in VIABLE_FLAGS:
            raise RuntimeError(
                f"{fn} must not have any arguments with the following names: {VIABLE_FLAGS}"
            )

cdef inline bint _run_sync(object function, dict kwargs):
    """
    Determines whether to run the function synchronously or asynchronously.

    This method checks for a flag in the kwargs and defers to it if present.
    If no flag is specified, it defers to the default execution mode.

    Args:
        kwargs: The keyword arguments passed to the function.

    Returns:
        True if the function should run synchronously, otherwise False.

    See Also:
        - :func:`_kwargs.get_flag_name`
    """
    cdef str flag = get_flag_name(kwargs)
    if flag:
        # If a flag was specified in the kwargs, we will defer to it.
        return is_sync(flag, kwargs, pop_flag=True)
    else:
        # No flag specified in the kwargs, we will defer to 'default'.
        return function._sync_default
    
cdef class _ASyncFunction(_ModifiedMixin):
    """
    A callable wrapper object that can be executed both synchronously and asynchronously.

    This class wraps a function or coroutine function, allowing it to be called in both
    synchronous and asynchronous contexts. It provides a flexible interface for handling
    different execution modes and applying various modifiers to the function's behavior.

    The class supports various modifiers that can alter the behavior of the function,
    such as caching, rate limiting, and execution in specific contexts (e.g., thread pools).

    Note:
        The logic for determining whether to execute the function synchronously or asynchronously
        is handled by the `self.fn` property, which checks for flags in the `kwargs` and defers
        to the default execution mode if no flags are specified.

    Example:
        async def my_coroutine(x: int) -> str:
            return str(x)

        func = ASyncFunction(my_coroutine)

        # Synchronous call
        result = func(5, sync=True)  # returns "5"

        # Asynchronous call
        result = await func(5)  # returns "5"

    See Also:
        - :class:`_ModifiedMixin`
        - :class:`ModifierManager`
    """

    def __init__(
        self,
        fn: AnyFn[P, T],
        _skip_validate: bint = False,
        **modifiers: Unpack[ModifierKwargs],
    ) -> None:
        """
        Initializes an ASyncFunction instance.

        Args:
            fn: The function to wrap.
            _skip_validate: For internal use only. Skips validation of the wrapped function when its already been validated once before.
            **modifiers: Keyword arguments for function modifiers.

        See Also:
            - :func:`_validate_wrapped_fn`
            - :class:`ModifierManager`
        """
        if not _skip_validate:
            _validate_wrapped_fn(fn)

        self.modifiers = ModifierManager(modifiers)
        """A :class:`~ModifierManager` instance managing function modifiers."""

        self._fn = None
        """The wrapped callable that will be called. This will be populated the first time it is needed."""

        self.__wrapped__ = fn
        """The original function that was wrapped."""

        self.__sync_default_cached = False
        self.__async_def_cached = False

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> MaybeCoro[T]:
        """
        Calls the wrapped function either synchronously or asynchronously.

        This method determines whether to execute the wrapped function synchronously
        or asynchronously based on the default mode and any provided flags. The
        decision logic is encapsulated within the `self.fn` property, which uses
        the `_run_sync` method to decide the execution mode.

        Note:
            The `self.fn` property is responsible for selecting the appropriate
            execution path (sync or async) by leveraging the `_run_sync` method.

        Args:
            *args: Positional arguments to pass to the wrapped function.
            **kwargs: Keyword arguments to pass to the wrapped function.

        Raises:
            Exception: Any exception that may be raised by the wrapped function.

        See Also:
            - :attr:`default`
            - :meth:`_run_sync`
        """
        _logger_debug(
            "calling %s fn: %s with args: %s kwargs: %s", self, self.fn, args, kwargs
        )
        return self.fn(*args, **kwargs)

    def __repr__(self) -> str:
        return "<{} {}.{} at {}>".format(
            self.__class__.__name__, self.__module__, self.__name__, hex(id(self))
        )

    @property
    def fn(self):
        # NOTE type hint doesnt work in py3.8 or py3.9, debug later
        #  -> Union[SyncFn[[CoroFn[P, T]], MaybeAwaitable[T]], SyncFn[[SyncFn[P, T]], MaybeAwaitable[T]]]:
        """
        Returns the final wrapped version of :attr:`ASyncFunction.__wrapped__` decorated with all of the a_sync goodness.

        Returns:
            The final wrapped function.

        See Also:
            - :meth:`_async_wrap`
            - :meth:`_sync_wrap`
        """
        fn = self._fn
        if fn is None:
            fn = self._async_wrap if self._async_def else self._sync_wrap
            self._fn = fn
        return fn

    if sys.version_info >= (3, 11) or TYPE_CHECKING:
        # we can specify P.args in python>=3.11 but in lower versions it causes a crash. Everything should still type check correctly on all versions.
        def map(
            self,
            *iterables: AnyIterable[P.args],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> "TaskMapping[P, T]":
            """
            Creates a TaskMapping for the wrapped function with the given iterables.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                A TaskMapping object for managing concurrent execution.

            See Also:
                - :class:`TaskMapping`
            """
            if TaskMapping is None:
                _import_TaskMapping()

            return TaskMapping(
                self,
                *iterables,
                concurrency=concurrency,
                name=task_name,
                **function_kwargs,
            )

        async def any(
            self,
            *iterables: AnyIterable[P.args],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> bint:
            """
            Checks if any result of the function applied to the iterables is truthy.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                True if any result is truthy, otherwise False.

            See Also:
                - :meth:`map`
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).any(pop=True, sync=False)

        async def all(
            self,
            *iterables: AnyIterable[P.args],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> bint:
            """
            Checks if all results of the function applied to the iterables are truthy.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                True if all results are truthy, otherwise False.

            See Also:
                - :meth:`map`
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).all(pop=True, sync=False)

        async def min(
            self,
            *iterables: AnyIterable[P.args],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> T:
            """
            Finds the minimum result of the function applied to the iterables.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                The minimum result.

            See Also:
                - :meth:`map`
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).min(pop=True, sync=False)

        async def max(
            self,
            *iterables: AnyIterable[P.args],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> T:
            """
            Finds the maximum result of the function applied to the iterables.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                The maximum result.

            See Also:
                - :meth:`map`
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).max(pop=True, sync=False)

        async def sum(
            self,
            *iterables: AnyIterable[P.args],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> T:
            """
            Calculates the sum of the results of the function applied to the iterables.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                The sum of the results.

            See Also:
                - :meth:`map`
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).sum(pop=True, sync=False)

    else:

        def map(
            self,
            *iterables: AnyIterable[Any],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> "TaskMapping[P, T]":
            """
            Creates a TaskMapping for the wrapped function with the given iterables.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                A TaskMapping object for managing concurrent execution.

            See Also:
                - :class:`TaskMapping`
            """
            if TaskMapping is None:
                _import_TaskMapping()

            return TaskMapping(
                self,
                *iterables,
                concurrency=concurrency,
                name=task_name,
                **function_kwargs,
            )

        async def any(
            self,
            *iterables: AnyIterable[Any],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> bint:
            """
            Checks if any result of the function applied to the iterables is truthy.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                True if any result is truthy, otherwise False.

            See Also:
                - :meth:`map`
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).any(pop=True, sync=False)

        async def all(
            self,
            *iterables: AnyIterable[Any],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> bint:
            """
            Checks if all results of the function applied to the iterables are truthy.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                True if all results are truthy, otherwise False.

            See Also:
                - :meth:`map`
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).all(pop=True, sync=False)

        async def min(
            self,
            *iterables: AnyIterable[Any],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> T:
            """
            Finds the minimum result of the function applied to the iterables.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                The minimum result.

            See Also:
                - :meth:`map`
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).min(pop=True, sync=False)

        async def max(
            self,
            *iterables: AnyIterable[Any],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> T:
            """
            Finds the maximum result of the function applied to the iterables.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                The maximum result.

            See Also:
                - :meth:`map`
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).max(pop=True, sync=False)

        async def sum(
            self,
            *iterables: AnyIterable[Any],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> T:
            """
            Calculates the sum of the results of the function applied to the iterables.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                The sum of the results.

            See Also:
                - :meth:`map`
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).sum(pop=True, sync=False)

    @property
    def _sync_default(self) -> bint:
        """
        Determines the default execution mode (sync or async) for the function.

        If the user did not specify a default, this method defers to the function's
        definition (sync vs async def).

        Returns:
            True if the default is sync, False if async.

        See Also:
            - :attr:`default`
        """
        if self.__sync_default_cached:
            return self.__sync_default

        cdef str default = self.get_default()
        cdef bint sync_default = (
            True
            if default == "sync"
            else False if default == "async" else not self._async_def
        )
        self.__sync_default = sync_default
        self.__sync_default_cached = True
        return sync_default

    @property
    def _async_def(self) -> bint:
        """
        Checks if the wrapped function is an asynchronous function.

        Returns:
            True if the function is asynchronous, otherwise False.

        See Also:
            - :func:`asyncio.iscoroutinefunction`
        """
        if self.__async_def_cached:
            return self.__async_def
        cdef bint async_def
        async_def = self.__async_def = iscoroutinefunction(self.__wrapped__)
        self.__async_def_cached = True
        return async_def

    @property
    def _asyncified(self) -> CoroFn[P, T]:
        """
        Converts the wrapped function to an asynchronous function and applies both sync and async modifiers.

        Raises:
            TypeError: If the wrapped function is already asynchronous.

        Returns:
            The asynchronous version of the wrapped function.

        See Also:
            - :meth:`_asyncify`
        """
        asyncified = self.__asyncified
        if asyncified is None:
            if self._async_def:
                raise TypeError(
                    f"Can only be applied to sync functions, not {self.__wrapped__}"
                )
            
            asyncified = self.__asyncified = self._asyncify(self._modified_fn)
        return asyncified

    @property
    def _modified_fn(self) -> AnyFn[P, T]:
        """
        Applies modifiers to the wrapped function.

        If the wrapped function is an asynchronous function, this method applies async modifiers.
        If the wrapped function is a synchronous function, this method applies sync modifiers.

        Returns:
            The modified function.

        See Also:
            - :meth:`ModifierManager.apply_async_modifiers`
            - :meth:`ModifierManager.apply_sync_modifiers`
        """
        modified_fn = self.__modified_fn
        if modified_fn is None:
            if self._async_def:
                modified_fn = self.__modified_fn = self.modifiers.apply_async_modifiers(self.__wrapped__)
            else:
                modified_fn = self.__modified_fn = self.modifiers.apply_sync_modifiers(self.__wrapped__)
        return modified_fn

    @property
    def _async_wrap(self):  # -> SyncFn[[CoroFn[P, T]], MaybeAwaitable[T]]:
        """
        The final wrapper if the wrapped function is an asynchronous function.

        This method applies the appropriate modifiers and determines whether to await the result.

        Returns:
            The wrapped function with async handling.

        See Also:
            - :meth:`_run_sync`
            - :meth:`_await`
        """
        async_wrap = self.__async_wrap
        if async_wrap is None:

            modified_fn = self._modified_fn
            await_helper = self.get_await()

            @wraps(modified_fn)
            def async_wrap(*args: P.args, **kwargs: P.kwargs) -> MaybeAwaitable[T]:  # type: ignore [name-defined]
                # sourcery skip: assign-if-exp
                # we dont want this so profiler outputs are more useful

                # Must take place before coro is created, we're popping a kwarg.
                should_await = _run_sync(self, kwargs)
                coro = modified_fn(*args, **kwargs)
                if should_await:
                    return await_helper(coro)
                else:
                    return coro
        
            self.__async_wrap = async_wrap

        return async_wrap

    @property
    def _sync_wrap(self):  # -> SyncFn[[SyncFn[P, T]], MaybeAwaitable[T]]:
        """
        The final wrapper if the wrapped function is a synchronous function.

        This method applies the appropriate modifiers and determines whether to run the function synchronously or asynchronously.

        Returns:
            The wrapped function with sync handling.

        See Also:
            - :meth:`_run_sync`
            - :meth:`_asyncified`
        """
        sync_wrap = self.__sync_wrap
        if sync_wrap is None:

            modified_fn = self._modified_fn
            asyncified = self._asyncified

            @wraps(modified_fn)
            def sync_wrap(*args: P.args, **kwargs: P.kwargs) -> MaybeAwaitable[T]:  # type: ignore [name-defined]
                if _run_sync(self, kwargs):
                    return modified_fn(*args, **kwargs)
                return asyncified(*args, **kwargs)
            
            self.__sync_wrap = sync_wrap

        return sync_wrap


class ASyncFunction(_ASyncFunction, Generic[P, T]):

    @overload
    def __init__(self, fn: CoroFn[P, T], **modifiers: Unpack[ModifierKwargs]) -> None:
        """
        Initializes an ASyncFunction instance for a coroutine function.

        Args:
            fn: The coroutine function to wrap.
            **modifiers: Keyword arguments for function modifiers.

        Example:
            async def my_coroutine(x: int) -> str:
                return str(x)

            func = ASyncFunction(my_coroutine, cache_type='memory')
        """

    @overload
    def __init__(self, fn: SyncFn[P, T], **modifiers: Unpack[ModifierKwargs]) -> None:
        """
        Initializes an ASyncFunction instance for a synchronous function.

        Args:
            fn: The synchronous function to wrap.
            **modifiers: Keyword arguments for function modifiers.

        Example:
            def my_function(x: int) -> str:
                return str(x)

            func = ASyncFunction(my_function, runs_per_minute=60)
        """
    
    def __init__(
        self,
        fn: AnyFn[P, T],
        _skip_validate: bint = False,
        **modifiers: Unpack[ModifierKwargs],
    ) -> None:
        _ASyncFunction.__init__(self, fn, _skip_validate=_skip_validate, **modifiers)
        update_wrapper(self, fn)
        if self.__doc__ is None:
            self.__doc__ = f"Since `{self.__name__}` is an {self.__docstring_append__}"
        else:
            self.__doc__ += (
                f"\n\nSince `{self.__name__}` is an {self.__docstring_append__}"
            )
    
    __docstring_append__ = ":class:`~a_sync.a_sync.function.ASyncFunction`, you can optionally pass either a `sync` or `asynchronous` kwarg with a boolean value."

    @overload
    def __call__(self, *args: P.args, sync: Literal[True], **kwargs: P.kwargs) -> T:
        """
        Calls the wrapped function synchronously.

        Args:
            *args: Positional arguments to pass to the wrapped function.
            sync: Must be True to indicate synchronous execution.
            **kwargs: Keyword arguments to pass to the wrapped function.

        Example:
            result = func(5, sync=True)
        """

    @overload
    def __call__(
        self, *args: P.args, sync: Literal[False], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]:
        """
        Calls the wrapped function asynchronously.

        Args:
            *args: Positional arguments to pass to the wrapped function.
            sync: Must be False to indicate asynchronous execution.
            **kwargs: Keyword arguments to pass to the wrapped function.

        Example:
            result = await func(5, sync=False)
        """

    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[False], **kwargs: P.kwargs
    ) -> T:
        """
        Calls the wrapped function synchronously.

        Args:
            *args: Positional arguments to pass to the wrapped function.
            asynchronous: Must be False to indicate synchronous execution.
            **kwargs: Keyword arguments to pass to the wrapped function.

        Example:
            result = func(5, asynchronous=False)
        """

    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[True], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]:
        """
        Calls the wrapped function asynchronously.

        Args:
            *args: Positional arguments to pass to the wrapped function.
            asynchronous: Must be True to indicate asynchronous execution.
            **kwargs: Keyword arguments to pass to the wrapped function.

        Example:
            result = await func(5, asynchronous=True)
        """

    @overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> MaybeCoro[T]:
        """
        Calls the wrapped function using the default execution mode.

        Args:
            *args: Positional arguments to pass to the wrapped function.
            **kwargs: Keyword arguments to pass to the wrapped function.

        Example:
            result = func(5)
        """


if sys.version_info < (3, 10):
    _inherit = ASyncFunction[AnyFn[P, T], ASyncFunction[P, T]]
else:
    _inherit = ASyncFunction[[AnyFn[P, T]], ASyncFunction[P, T]]


cdef class ASyncDecorator(_ModifiedMixin):
    def __cinit__(self, **modifiers: Unpack[ModifierKwargs]) -> None:
        """
        Initializes an ASyncDecorator instance by validating the inputs.

        Args:
            **modifiers: Keyword arguments for function modifiers.

        Raises:
            ValueError: If 'default' is not 'sync', 'async', or None.

        See Also:
            - :class:`ModifierManager`
        """
        if modifiers.get("default", object) not in ("sync", "async", None):
            if "default" not in modifiers:
                raise ValueError("you must pass a value for `default`")
            raise ValueError(
                f"'default' must be either 'sync', 'async', or None. You passed {modifiers['default']}."
            )
        self.modifiers = ModifierManager(modifiers)

    @overload
    def __call__(self, func: AnyFn[Concatenate[B, P], T]) -> "ASyncBoundMethod[B, P, T]":  # type: ignore [override]
        """
        Decorates a bound method with async or sync behavior based on the default modifier.

        Args:
            func: The bound method to decorate.

        Returns:
            An ASyncBoundMethod instance with the appropriate default behavior.

        See Also:
            - :class:`ASyncBoundMethod`
        """

    @overload
    def __call__(self, func: AnyFn[P, T]) -> ASyncFunction[P, T]:  # type: ignore [override]
        """
        Decorates a function with async or sync behavior based on the default modifier.

        Args:
            func: The function to decorate.

        Returns:
            An ASyncFunction instance with the appropriate default behavior.

        See Also:
            - :class:`ASyncFunction`
        """

    def __call__(self, func: AnyFn[P, T]) -> ASyncFunction[P, T]:  # type: ignore [override]
        """
        Decorates a function with async or sync behavior based on the default modifier.

        Args:
            func: The function to decorate.

        Returns:
            An ASyncFunction instance with the appropriate default behavior.

        See Also:
            - :class:`ASyncFunction`
        """
        cdef str default = self.get_default()
        if default == "async":
            return ASyncFunctionAsyncDefault(func, **self.modifiers._modifiers)
        elif default == "sync":
            return ASyncFunctionSyncDefault(func, **self.modifiers._modifiers)
        elif iscoroutinefunction(func):
            return ASyncFunctionAsyncDefault(func, **self.modifiers._modifiers)
        else:
            return ASyncFunctionSyncDefault(func, **self.modifiers._modifiers)


cdef set[uintptr_t] _is_genfunc_cache = set()

cdef void _check_not_genfunc_cached(func: Callable):
    cdef uintptr_t fid = id(func)
    if fid not in _is_genfunc_cache:
        _check_not_genfunc(func)
        _is_genfunc_cache.add(fid)

cdef void _check_not_genfunc(func: Callable):
    """Raises an error if the function is a generator or async generator.

    Args:
        func: The function to check.

    Raises:
        ValueError: If the function is a generator or async generator.

    See Also:
        - :func:`inspect.isasyncgenfunction`
        - :func:`inspect.isgeneratorfunction`
    """
    if isasyncgenfunction(func) or isgeneratorfunction(func):
        raise ValueError("unable to decorate generator functions with this decorator")


# Mypy helper classes


class ASyncFunctionSyncDefault(ASyncFunction[P, T]):
    """A specialized :class:`~ASyncFunction` that defaults to synchronous execution.

    This class is used when the :func:`~a_sync` decorator is applied with `default='sync'`.
    It provides type hints to indicate that the default call behavior is synchronous and
    supports IDE type checking for most use cases.

    The wrapped function can still be called asynchronously by passing `sync=False`
    or `asynchronous=True` as a keyword argument.

    Example:
        @a_sync(default='sync')
        async def my_function(x: int) -> str:
            return str(x)

        # Synchronous call (default behavior)
        result = my_function(5)  # returns "5"

        # Asynchronous call
        result = await my_function(5, sync=False)  # returns "5"
    """

    @overload
    def __call__(self, *args: P.args, sync: Literal[True], **kwargs: P.kwargs) -> T:
        # TODO write specific docs for this overload
        ...

    @overload
    def __call__(
        self, *args: P.args, sync: Literal[False], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]:
        # TODO write specific docs for this overload
        ...

    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[False], **kwargs: P.kwargs
    ) -> T:
        # TODO write specific docs for this overload
        ...

    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[True], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]:
        # TODO write specific docs for this overload
        ...

    @overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        # TODO write specific docs for this overload
        ...

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> MaybeCoro[T]:
        """Calls the wrapped function, defaulting to synchronous execution.

        This method overrides the base :meth:`ASyncFunction.__call__` to provide a synchronous
        default behavior.

        Args:
            *args: Positional arguments to pass to the wrapped function.
            **kwargs: Keyword arguments to pass to the wrapped function.

        Raises:
            Exception: Any exception that may be raised by the wrapped function.

        Returns:
            The result of the function call.

        See Also:
            - :meth:`ASyncFunction.__call__`
        """
        return self.fn(*args, **kwargs)

    __docstring_append__ = ":class:`~a_sync.a_sync.function.ASyncFunctionSyncDefault`, you can optionally pass `sync=False` or `asynchronous=True` to force it to return a coroutine. Without either kwarg, it will run synchronously."


class ASyncFunctionAsyncDefault(ASyncFunction[P, T]):
    """
    A specialized :class:`~ASyncFunction` that defaults to asynchronous execution.

    This class is used when the :func:`~a_sync` decorator is applied with `default='async'`.
    It provides type hints to indicate that the default call behavior is asynchronous
    and supports IDE type checking for most use cases.

    The wrapped function can still be called synchronously by passing `sync=True`
    or `asynchronous=False` as a keyword argument.

    Example:
        @a_sync(default='async')
        async def my_function(x: int) -> str:
            return str(x)

        # Asynchronous call (default behavior)
        result = await my_function(5)  # returns "5"

        # Synchronous call
        result = my_function(5, sync=True)  # returns "5"
    """

    @overload
    def __call__(self, *args: P.args, sync: Literal[True], **kwargs: P.kwargs) -> T:
        # TODO write specific docs for this overload
        ...

    @overload
    def __call__(
        self, *args: P.args, sync: Literal[False], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]:
        # TODO write specific docs for this overload
        ...

    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[False], **kwargs: P.kwargs
    ) -> T:
        # TODO write specific docs for this overload
        ...

    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[True], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]:
        # TODO write specific docs for this overload
        ...

    @overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Coroutine[Any, Any, T]: ...
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> MaybeCoro[T]:
        """Calls the wrapped function, defaulting to asynchronous execution.

        This method overrides the base :meth:`ASyncFunction.__call__` to provide an asynchronous
        default behavior.

        Args:
            *args: Positional arguments to pass to the wrapped function.
            **kwargs: Keyword arguments to pass to the wrapped function.

        Raises:
            Exception: Any exception that may be raised by the wrapped function.

        Returns:
            The result of the function call.

        See Also:
            - :meth:`ASyncFunction.__call__`
        """
        return self.fn(*args, **kwargs)

    __docstring_append__ = ":class:`~a_sync.a_sync.function.ASyncFunctionAsyncDefault`, you can optionally pass `sync=True` or `asynchronous=False` to force it to run synchronously and return a value. Without either kwarg, it will return a coroutine for you to await."


cdef class ASyncDecoratorSyncDefault(ASyncDecorator):
    @overload
    def __call__(self, func: AnyFn[Concatenate[B, P], T]) -> "ASyncBoundMethodSyncDefault[P, T]":  # type: ignore [override]
        """
        Decorates a bound method with synchronous default behavior.

        Args:
            func: The bound method to decorate.

        Returns:
            An ASyncBoundMethodSyncDefault instance with synchronous default behavior.

        See Also:
            - :class:`ASyncBoundMethodSyncDefault`
        """

    @overload
    def __call__(self, func: AnyBoundMethod[P, T]) -> ASyncFunctionSyncDefault[P, T]:  # type: ignore [override]
        """
        Decorates a bound method with synchronous default behavior.

        Args:
            func: The bound method to decorate.

        Returns:
            An ASyncFunctionSyncDefault instance with synchronous default behavior.

        See Also:
            - :class:`ASyncFunctionSyncDefault`
        """

    @overload
    def __call__(self, func: AnyFn[P, T]) -> ASyncFunctionSyncDefault[P, T]:  # type: ignore [override]
        """
        Decorates a function with synchronous default behavior.

        Args:
            func: The function to decorate.

        Returns:
            An ASyncFunctionSyncDefault instance with synchronous default behavior.

        See Also:
            - :class:`ASyncFunctionSyncDefault`
        """

    def __call__(self, func: AnyFn[P, T]) -> ASyncFunctionSyncDefault[P, T]:
        # TODO write specific docs for this implementation
        return ASyncFunctionSyncDefault(func, **self.modifiers._modifiers)


cdef class ASyncDecoratorAsyncDefault(ASyncDecorator):
    @overload
    def __call__(self, func: AnyFn[Concatenate[B, P], T]) -> "ASyncBoundMethodAsyncDefault[P, T]":  # type: ignore [override]
        """
        Decorates a bound method with asynchronous default behavior.

        Args:
            func: The bound method to decorate.

        Returns:
            An ASyncBoundMethodAsyncDefault instance with asynchronous default behavior.

        See Also:
            - :class:`ASyncBoundMethodAsyncDefault`
        """

    @overload
    def __call__(self, func: AnyBoundMethod[P, T]) -> ASyncFunctionAsyncDefault[P, T]:  # type: ignore [override]
        """
        Decorates a bound method with asynchronous default behavior.

        Args:
            func: The bound method to decorate.

        Returns:
            An ASyncFunctionAsyncDefault instance with asynchronous default behavior.

        See Also:
            - :class:`ASyncFunctionAsyncDefault`
        """

    @overload
    def __call__(self, func: AnyFn[P, T]) -> ASyncFunctionAsyncDefault[P, T]:  # type: ignore [override]
        """
        Decorates a function with asynchronous default behavior.

        Args:
            func: The function to decorate.

        Returns:
            An ASyncFunctionAsyncDefault instance with asynchronous default behavior.

        See Also:
            - :class:`ASyncFunctionAsyncDefault`
        """

    def __call__(self, func: AnyFn[P, T]) -> ASyncFunctionAsyncDefault[P, T]:
        # TODO write specific docs for this implementation
        return ASyncFunctionAsyncDefault(func, **self.modifiers._modifiers)


cdef inline void _import_TaskMapping():
    global TaskMapping
    from a_sync import TaskMapping



from asyncio import Lock
from typing import Any, Callable, Dict, Generic, Optional, TypeVar, Union

from a_sync.async_property.proxy import AwaitableOnly, AwaitableProxy

ProxyType = Union[AwaitableOnly[__T], AwaitableProxy[__T]]

FieldName = str

__I = TypeVar("__I")
__T = TypeVar("__T")

def async_cached_property(
    func: Callable[[__I], __T], *args, **kwargs
) -> "AsyncCachedPropertyDescriptor[__I, __T]": ...

class AsyncCachedPropertyInstanceState(Generic[__T]):
    cache: Dict[FieldName, __T]
    locks: Dict[FieldName, Lock]

class AsyncCachedPropertyDescriptor(Generic[__I, __T]):
    field_name: FieldName
    _load_value: Callable[[__I], __T]
    _fget: Callable[[__I], __T]
    _fset: Optional[Callable]
    _fdel: Optional[Callable]
    def __init__(
        self, _fget: Callable[[__I], __T], _fset=None, _fdel=None, field_name=None
    ) -> None: ...
    def __set_name__(self, owner, name) -> None: ...
    def __get__(self, instance: __I, owner) -> ProxyType[__T]: ...
    def __set__(self, instance: __I, value): ...
    def __delete__(self, instance: __I) -> None: ...
    def setter(self, method): ...
    def deleter(self, method): ...
    def _check_method_name(self, method, method_type) -> None: ...
    def _check_method_sync(self, method, method_type) -> None: ...
    def get_instance_state(self, instance) -> AsyncCachedPropertyInstanceState[__T]: ...
    def get_lock(self, instance: __I) -> Lock: ...
    def get_cache(self, instance: __I) -> Dict[field_name, Any]: ...
    def has_cache_value(self, instance: __I) -> bool: ...
    def get_cache_value(self, instance: __I) -> __T: ...
    def set_cache_value(self, instance: __I, value: __T) -> None: ...
    def del_cache_value(self, instance: __I) -> None: ...
    def get_loader(self, instance: __I) -> Callable[[], __T]: ...



import os
from glob import glob
from pathlib import Path
from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup


with open("requirements.txt", "r") as f:
    requirements = list(map(str.strip, f.read().split("\n")))[:-1]

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="ez_a_sync",
    packages=find_packages(exclude=["tests", "tests.*"]),
    use_scm_version={
        "root": ".",
        "relative_to": __file__,
        "local_scheme": "no-local-version",
        "version_scheme": "python-simplified-semver",
    },
    description="A library that makes it easy to define objects that can be used for both sync and async use cases.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="BobTheBuidler",
    author_email="bobthebuidlerdefi@gmail.com",
    url="https://github.com/BobTheBuidler/a-sync",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
    ],
    install_requires=requirements,
    setup_requires=["setuptools_scm", "cython"],
    python_requires=">=3.8,<3.14",
    package_data={
        "a_sync": ["py.typed", "*.pxd", "**/*.pxd"],
    },
    include_package_data=True,
    ext_modules=cythonize(
        [
            Extension(
                os.path.splitext(pyx_path)[0].replace(os.path.sep, "."),
                sources=[pyx_path],
                include_dirs=["include"],
            )
            for pyx_path in glob("a_sync/**/*.pyx", recursive=True)
        ],
        compiler_directives={
            "language_level": 3,
            "embedsignature": True,
            "linetrace": True,
        },
    ),
    zip_safe=False,
)



import asyncio
import functools
import typing
from copy import copy as _copy
from logging import getLogger
from typing import Any, Awaitable, Callable, Generator, Optional, Tuple, Type, Union

import async_property as ap  # type: ignore [import]
from typing_extensions import Concatenate, Self, Unpack

from a_sync._smart cimport shield
from a_sync._typing import (AnyFn, AnyGetterFunction, AnyIterable, AsyncGetterFunction, 
                            DefaultMode, I, ModifierKwargs, P, T)
from a_sync.a_sync import _descriptor, config, function, method
from a_sync.a_sync._helpers cimport _asyncify, _await

from a_sync.a_sync.function cimport _ModifiedMixin
from a_sync.a_sync.method cimport _is_a_sync_instance, _update_cache_timer
from a_sync.async_property import cached
from a_sync.async_property.cached cimport AsyncCachedPropertyInstanceState
from a_sync.asyncio.create_task cimport ccreate_task_simple
from a_sync.functools cimport wraps

if typing.TYPE_CHECKING:
    from a_sync.task import TaskMapping

else:
    # to prevent circ import issues we will import this later
    TaskMapping = None


# cdef asyncio
cdef object get_event_loop = asyncio.get_event_loop
cdef object iscoroutinefunction = asyncio.iscoroutinefunction
cdef object Lock = asyncio.Lock
cdef object Task = asyncio.Task
del asyncio

# cdef copy
cdef object copy = _copy
del _copy

# cdef functools
cdef object partial = functools.partial
del functools

# cdef logging
cdef public object logger = getLogger(__name__)
cdef object _logger_is_enabled = logger.isEnabledFor
cdef object _logger_debug = logger.debug
cdef object _logger_log = logger._log
cdef object DEBUG = 10
del getLogger

# cdef typing
cdef object Literal = typing.Literal
cdef object overload = typing.overload
del typing


# cdef a_sync.a_sync._descriptor
cdef object ASyncDescriptor = _descriptor.ASyncDescriptor
del _descriptor

# cdef a_sync.a_sync.function
cdef object ASyncFunction = function.ASyncFunction
cdef object ASyncFunctionAsyncDefault = function.ASyncFunctionAsyncDefault
cdef object ASyncFunctionSyncDefault = function.ASyncFunctionSyncDefault
del function

# cdef a_sync.a_sync.method
cdef object ASyncBoundMethod = method.ASyncBoundMethod
cdef object ASyncBoundMethodAsyncDefault = method.ASyncBoundMethodAsyncDefault
cdef object ASyncMethodDescriptorAsyncDefault = method.ASyncMethodDescriptorAsyncDefault
del method


class _ASyncPropertyDescriptorBase(ASyncDescriptor[I, Tuple[()], T]):
    """Base class for creating asynchronous properties.

    This class provides the foundation for defining properties that can be accessed
    both synchronously and asynchronously. It includes utility methods for common
    operations such as `any`, `all`, `min`, `max`, and `sum`.
    """

    any: ASyncFunction[AnyIterable[I], bool]
    """An ASyncFunction that checks if any result is truthy."""

    all: ASyncFunction[AnyIterable[I], bool]
    """An ASyncFunction that checks if all results are truthy."""

    min: ASyncFunction[AnyIterable[I], T]
    """An ASyncFunction that returns the minimum result."""

    max: ASyncFunction[AnyIterable[I], T]
    """An ASyncFunction that returns the maximum result."""

    sum: ASyncFunction[AnyIterable[I], T]
    """An ASyncFunction that returns the sum of results."""

    hidden_method_descriptor: "HiddenMethodDescriptor[T]"
    """A descriptor for the hidden method."""

    __wrapped__: Callable[[I], T]
    """The wrapped function or method."""

    __slots__ = "hidden_method_name", "hidden_method_descriptor", "_fget"

    def __init__(
        _ModifiedMixin self,
        _fget: AsyncGetterFunction[I, T],
        field_name: Optional[str] = None,
        **modifiers: Unpack[ModifierKwargs],
    ) -> None:
        """Initializes the _ASyncPropertyDescriptorBase.

        Args:
            _fget: The function to be wrapped.
            field_name: Optional name for the field. If not provided, the function's name will be used.
            **modifiers: Additional modifier arguments.
        """
        cdef dict hidden_modifiers
        ASyncDescriptor.__init__(self, _fget, field_name, **modifiers)
        self.hidden_method_name = f"__{self.field_name}__"
        hidden_modifiers = self.modifiers._modifiers.copy()
        hidden_modifiers["default"] = "async"
        self.hidden_method_descriptor = HiddenMethodDescriptor(
            self.get, self.hidden_method_name, **hidden_modifiers
        )
        if iscoroutinefunction(_fget):
            self._fget = self.__wrapped__
        else:
            self._fget = _asyncify(self.__wrapped__, self.modifiers.executor)

    @overload
    def __get__(self, instance: None, owner: Type[I]) -> Self: ...
    @overload
    def __get__(self, instance: I, owner: Type[I]) -> Awaitable[T]: ...
    def __get__(
        self, instance: Optional[I], owner: Type[I]
    ) -> Union[Self, Awaitable[T]]:
        """Retrieves the property value, either synchronously or asynchronously.

        Args:
            instance: The instance from which the property is accessed.
            owner: The owner class of the property.

        Returns:
            The property value, either as an awaitable or directly.
        """
        if instance is None:
            return self
        
        cdef object awaitable = super().__get__(instance, owner)

        # if the user didn't specify a default behavior, we will defer to the instance
        cdef bint should_await
        cdef str default = _ModifiedMixin.get_default(self)
        if default:
            should_await = default == "sync"
        elif _is_a_sync_instance(instance):
            should_await = instance.__a_sync_instance_should_await__
        else:
            should_await = not get_event_loop().is_running()
        
        cdef object retval
        cdef bint debug_logs = _logger_is_enabled(DEBUG)
        if should_await:
            if debug_logs:
                _logger_log(
                    DEBUG,
                    "awaiting awaitable for %s for instance: %s owner: %s",
                    (awaitable, self, instance, owner),
                )
            retval = _await(awaitable)
        else:
            retval = awaitable

        if debug_logs:
            _logger_log(
                DEBUG,
                "returning %s for %s for instance: %s owner: %s",
                (retval, self, instance, owner), 
            )

        return retval

    async def get(self, instance: I, owner: Optional[Type[I]] = None) -> T:
        """Asynchronously retrieves the property value.

        Args:
            instance: The instance from which the property is accessed.
            owner: The owner class of the property.

        Returns:
            The property value.
        """
        _logger_debug("awaiting %s for instance %s", self, instance)
        return await super().__get__(instance, owner)

    def map(
        self,
        instances: AnyIterable[I],
        owner: Optional[Type[I]] = None,
        concurrency: Optional[int] = None,
        name: str = "",
    ) -> "TaskMapping[I, T]":
        """Maps the property across multiple instances.

        Args:
            instances: An iterable of instances.
            owner: The owner class of the property.
            concurrency: Optional concurrency limit.
            name: Optional name for the task mapping.

        Returns:
            A TaskMapping object.
        """
        _logger_debug("mapping %s to instances: %s owner: %s", self, instances, owner)
        
        """NOTE: This silly helper just fixes a circular import"""
        if TaskMapping is None:
            _import_TaskMapping()

        return TaskMapping(
            self,
            instances,
            owner=owner,
            name=name or self.field_name,
            concurrency=concurrency,
        )


class ASyncPropertyDescriptor(
    _ASyncPropertyDescriptorBase[I, T], ap.base.AsyncPropertyDescriptor
):
    """Descriptor class for asynchronous properties."""


class ASyncPropertyDescriptorSyncDefault(ASyncPropertyDescriptor[I, T]):
    """
    A variant of :class:`~ASyncPropertyDescriptor` that defaults to synchronous behavior.

    This class is used when the property is primarily intended to be accessed
    synchronously but can also be used asynchronously if needed.
    """

    # TODO give all of these docstrings
    default = "sync"
    # TODO and give these ones examples
    any: ASyncFunctionSyncDefault[AnyIterable[I], bool]
    all: ASyncFunctionSyncDefault[AnyIterable[I], bool]
    min: ASyncFunctionSyncDefault[AnyIterable[I], T]
    max: ASyncFunctionSyncDefault[AnyIterable[I], T]
    sum: ASyncFunctionSyncDefault[AnyIterable[I], T]

    @overload
    def __get__(self, instance: None, owner: Type[I]) -> Self: ...
    @overload
    def __get__(self, instance: I, owner: Type[I]) -> T: ...

    __get__ = _ASyncPropertyDescriptorBase.__get__
    """Retrieves the property value, either synchronously or asynchronously.

    Args:
        instance: The instance from which the property is accessed.
        owner: The owner class of the property.

    Returns:
        The property value, either as an awaitable or directly.
    """


class ASyncPropertyDescriptorAsyncDefault(ASyncPropertyDescriptor[I, T]):
    """
    A variant of :class:`~ASyncPropertyDescriptor` that defaults to asynchronous behavior.

    This class is used when the property is primarily intended to be accessed
    asynchronously but can also be used synchronously if needed.
    """

    # TODO give all of these docstrings
    default = "async"
    # TODO and give these ones examples
    any: ASyncFunctionAsyncDefault[AnyIterable[I], bool]
    all: ASyncFunctionAsyncDefault[AnyIterable[I], bool]
    min: ASyncFunctionAsyncDefault[AnyIterable[I], T]
    max: ASyncFunctionAsyncDefault[AnyIterable[I], T]
    sum: ASyncFunctionAsyncDefault[AnyIterable[I], T]


# Give all of these docstrings
ASyncPropertyDecorator = Callable[[AnyGetterFunction[I, T]], ASyncPropertyDescriptor[I, T]]
ASyncPropertyDecoratorSyncDefault = Callable[
    [AnyGetterFunction[I, T]], ASyncPropertyDescriptorSyncDefault[I, T]
]
ASyncPropertyDecoratorAsyncDefault = Callable[
    [AnyGetterFunction[I, T]], ASyncPropertyDescriptorAsyncDefault[I, T]
]


@overload
def a_sync_property(  # type: ignore [misc]
    func: Literal[None] = None,
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDecorator[I, T]: ...


@overload
def a_sync_property(  # type: ignore [misc]
    func: AnyGetterFunction[I, T],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDescriptor[I, T]: ...


@overload
def a_sync_property(  # type: ignore [misc]
    func: Literal[None],
    default: Literal["sync"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDecoratorSyncDefault[I, T]: ...


@overload
def a_sync_property(  # type: ignore [misc]
    func: Literal[None],
    default: Literal["sync"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDecoratorSyncDefault[I, T]: ...


@overload
def a_sync_property(  # type: ignore [misc]
    func: Literal[None],
    default: Literal["async"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDecoratorAsyncDefault[I, T]: ...


@overload
def a_sync_property(  # type: ignore [misc]
    func: Literal[None],
    default: DefaultMode = config.DEFAULT_MODE,
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDecorator[I, T]: ...


@overload
def a_sync_property(  # type: ignore [misc]
    default: Literal["sync"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDecoratorSyncDefault[I, T]: ...


@overload
def a_sync_property(  # type: ignore [misc]
    default: Literal["async"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDecoratorAsyncDefault[I, T]: ...


@overload
def a_sync_property(  # type: ignore [misc]
    func: AnyGetterFunction[I, T],
    default: Literal["sync"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDescriptorSyncDefault[I, T]: ...


@overload
def a_sync_property(  # type: ignore [misc]
    func: AnyGetterFunction[I, T],
    default: Literal["async"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDescriptorAsyncDefault[I, T]: ...


@overload
def a_sync_property(  # type: ignore [misc]
    func: AnyGetterFunction[I, T],
    default: DefaultMode = config.DEFAULT_MODE,
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDescriptor[I, T]: ...


def a_sync_property(  # type: ignore [misc]
    func: Union[AnyGetterFunction[I, T], DefaultMode] = None,
    **modifiers: Unpack[ModifierKwargs],
) -> Union[
    ASyncPropertyDescriptor[I, T],
    ASyncPropertyDescriptorSyncDefault[I, T],
    ASyncPropertyDescriptorAsyncDefault[I, T],
    ASyncPropertyDecorator[I, T],
    ASyncPropertyDecoratorSyncDefault[I, T],
    ASyncPropertyDecoratorAsyncDefault[I, T],
]:
    """Decorator for creating properties that can be accessed both synchronously and asynchronously.

    Args:
        func: The function to be wrapped.
        **modifiers: Additional modifier arguments.

    Returns:
        A property descriptor that supports both sync and async access.
    """
    func, modifiers = _parse_args(func, <dict>modifiers)
    cdef object descriptor_class
    if (<dict>modifiers).get("default") == "sync":
        descriptor_class = ASyncPropertyDescriptorSyncDefault
    elif (<dict>modifiers).get("default") == "async":
        descriptor_class = ASyncPropertyDescriptorAsyncDefault
    else:
        descriptor_class = ASyncPropertyDescriptor
    decorator = partial(descriptor_class, **modifiers)
    return decorator if func is None else decorator(func)


class ASyncCachedPropertyDescriptor(
    _ASyncPropertyDescriptorBase[I, T], cached.AsyncCachedPropertyDescriptor
):
    """
    A descriptor class for dual-function sync/async cached properties.

    This class extends the API of ASyncPropertyDescriptor to provide
    caching functionality, storing the computed value after the first access.
    """

    __slots__ = "_fset", "_fdel", "__async_property__"

    def __init__(
        self,
        _fget: AsyncGetterFunction[I, T],
        _fset=None,
        _fdel=None,
        field_name=None,
        **modifiers: Unpack[ModifierKwargs],
    ) -> None:
        """Initializes the ASyncCachedPropertyDescriptor.

        Args:
            _fget: The function to be wrapped.
            _fset: Optional setter function for the property.
            _fdel: Optional deleter function for the property.
            field_name: Optional name for the field. If not provided, the function's name will be used.
            **modifiers: Additional modifier arguments.
        """
        _ASyncPropertyDescriptorBase.__init__(self, _fget, field_name, **modifiers)
        self._check_method_sync(_fset, "setter")
        self._fset = _fset
        """Optional setter function for the property."""

        self._check_method_sync(_fdel, "deleter")
        self._fdel = _fdel
        """Optional deleter function for the property."""

    def get_lock(self, instance: I) -> "Task[T]":
        """Retrieves the lock for the property.

        Args:
            instance: The instance from which the property is accessed.

        Returns:
            An asyncio Task representing the lock.
        """
        locks = (<AsyncCachedPropertyInstanceState>self.get_instance_state(instance)).locks 
        task = locks[self.field_name]
        if isinstance(task, Lock):
            # default behavior uses lock but we want to use a Task so all waiters wake up together
            task = ccreate_task_simple(self._fget(instance))
            locks[self.field_name] = task
        return task

    def get_loader(self, instance: I) -> Callable[[], T]:
        """Retrieves the loader function for the property.

        Args:
            instance: The instance from which the property is accessed.

        Returns:
            A callable that loads the property value.
        """
        cdef str field_name

        loader = self._load_value
        if loader is None:
            field_name = self.field_name

            @wraps(self._fget)
            async def loader(instance):
                cdef AsyncCachedPropertyInstanceState cache_state
                cache_state = self.get_instance_state(instance)

                inner_task = cache_state.locks[field_name]
                if isinstance(inner_task, Lock):
                    # default behavior uses lock but we want to use a Task so all waiters wake up together
                    inner_task = ccreate_task_simple(self._fget(instance))
                    cache_state.locks[field_name] = inner_task

                try:
                    value = await shield(inner_task)
                except Exception as e:
                    instance_context = {"property": self, "instance": instance}
                    if e.args and e.args[-1] != instance_context:
                        e.args = *e.args, instance_context
                    raise copy(e).with_traceback(e.__traceback__)
                
                if self._fset is not None:
                    self._fset(instance, value)
                
                if field_name not in cache_state.cache:
                    cache_state.cache[field_name] = value
                    cache_state.locks.pop(field_name)
                
                return value

            self._load_value = loader

        return lambda: loader(instance)


class ASyncCachedPropertyDescriptorSyncDefault(ASyncCachedPropertyDescriptor[I, T]):
    """
    A variant of :class:`~ASyncCachedPropertyDescriptor` that defaults to synchronous behavior.

    This class is used for cached properties that are primarily intended to be
    accessed synchronously but can also be used asynchronously if needed.

    Note:
        You should never create these yourself. They are automatically generated by ez-a-sync internally.
    """

    default: Literal["sync"]

    @overload
    def __get__(self, instance: None, owner: Type[I]) -> Self: ...
    @overload
    def __get__(self, instance: I, owner: Type[I]) -> T: ...

    __get__ = _ASyncPropertyDescriptorBase.__get__
    """Retrieves the cached property value, either synchronously or asynchronously.

    Args:
        instance: The instance from which the property is accessed.
        owner: The owner class of the property.

    Returns:
        The cached property value, either as an awaitable or directly.
    """


class ASyncCachedPropertyDescriptorAsyncDefault(ASyncCachedPropertyDescriptor[I, T]):
    """
    A variant of :class:`~ASyncCachedPropertyDescriptor` that defaults to asynchronous behavior.

    This class is used for cached properties that are primarily intended to be
    accessed asynchronously but can also be used synchronously if needed.

    Note:
        You should never create these yourself. They are automatically generated by ez-a-sync internally.
    """

    default: Literal["async"]


ASyncCachedPropertyDecorator = Callable[
    [AnyGetterFunction[I, T]], ASyncCachedPropertyDescriptor[I, T]
]
ASyncCachedPropertyDecoratorSyncDefault = Callable[
    [AnyGetterFunction[I, T]], ASyncCachedPropertyDescriptorSyncDefault[I, T]
]
ASyncCachedPropertyDecoratorAsyncDefault = Callable[
    [AnyGetterFunction[I, T]], ASyncCachedPropertyDescriptorAsyncDefault[I, T]
]


@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: Literal[None] = None,
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDecorator[I, T]: ...


@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: AnyGetterFunction[I, T],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDescriptor[I, T]: ...


@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: Literal[None],
    default: Literal["sync"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDecoratorSyncDefault[I, T]: ...


@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: Literal[None],
    default: Literal["async"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDecoratorAsyncDefault[I, T]: ...


@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: Literal[None],
    default: DefaultMode,
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDecorator[I, T]: ...


@overload
def a_sync_cached_property(  # type: ignore [misc]
    default: Literal["sync"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDecoratorSyncDefault[I, T]: ...


@overload
def a_sync_cached_property(  # type: ignore [misc]
    default: Literal["async"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDecoratorAsyncDefault[I, T]: ...


@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: AnyGetterFunction[I, T],
    default: Literal["sync"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDescriptorSyncDefault[I, T]: ...


@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: AnyGetterFunction[I, T],
    default: Literal["async"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDescriptorAsyncDefault[I, T]: ...


@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: AnyGetterFunction[I, T],
    default: DefaultMode = config.DEFAULT_MODE,
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDescriptor[I, T]: ...


def a_sync_cached_property(  # type: ignore [misc]
    func: Optional[AnyGetterFunction[I, T]] = None,
    **modifiers: Unpack[ModifierKwargs],
) -> Union[
    ASyncCachedPropertyDescriptor[I, T],
    ASyncCachedPropertyDescriptorSyncDefault[I, T],
    ASyncCachedPropertyDescriptorAsyncDefault[I, T],
    ASyncCachedPropertyDecorator[I, T],
    ASyncCachedPropertyDecoratorSyncDefault[I, T],
    ASyncCachedPropertyDecoratorAsyncDefault[I, T],
]:
    """Decorator for creating cached properties that can be accessed both synchronously and asynchronously.

    Args:
        func: The function to be wrapped.
        **modifiers: Additional modifier arguments.

    Returns:
        A cached property descriptor that supports both sync and async access.
    """
    func, modifiers = _parse_args(func, <dict>modifiers)
    cdef object descriptor_class
    if (<dict>modifiers).get("default") == "sync":
        descriptor_class = ASyncCachedPropertyDescriptorSyncDefault
    elif (<dict>modifiers).get("default") == "async":
        descriptor_class = ASyncCachedPropertyDescriptorAsyncDefault
    else:
        descriptor_class = ASyncCachedPropertyDescriptor
    decorator = partial(descriptor_class, **modifiers)
    return decorator if func is None else decorator(func)


class HiddenMethod(ASyncBoundMethodAsyncDefault[I, Tuple[()], T]):
    """Represents a hidden method for asynchronous properties.

    This class is used internally to manage hidden getter methods associated with a/sync properties.

    Note:
        You should never create these yourself. They are automatically generated by ez-a-sync internally.
    """

    def __init__(
        self,
        instance: I,
        unbound: AnyFn[Concatenate[I, P], T],
        async_def: bool,
        field_name: str,
        **modifiers: Unpack[ModifierKwargs],
    ) -> None:
        """Initializes the HiddenMethod.

        Args:
            instance: The instance to which the method is bound.
            unbound: The unbound function to be wrapped.
            async_def: Indicates if the method is asynchronous.
            field_name: The name of the field associated with the method.
            **modifiers: Additional modifier arguments.
        """
        ASyncBoundMethod.__init__(self, instance, unbound, async_def, **modifiers)
        self.__name__ = field_name
        """The name of the hidden method."""

    def __repr__(self) -> str:
        """Returns a string representation of the HiddenMethod."""
        instance_type = type(self.__self__)
        return "<{} for property {}.{}.{} bound to {}>".format(
            self.__class__.__name__,
            instance_type.__module__,
            instance_type.__name__,
            self.__name__[2:-2],
            self.__self__,
        )

    def __await__(self) -> Generator[Any, None, T]:
        """Returns an awaitable for the method."""
        # NOTE: self(sync=False).__await__() would be cleaner but requires way more compute for no real gain
        _logger_debug("awaiting %s", self)
        return self.fn(self.__self__, sync=False).__await__()


class HiddenMethodDescriptor(ASyncMethodDescriptorAsyncDefault[I, Tuple[()], T]):
    """Descriptor for hidden methods associated with asynchronous properties.

    This class is used internally to manage hidden getter methods associated with a/sync properties.

    Note:
        You should never create these yourself. They are automatically generated by ez-a-sync internally.
    """

    def __init__(
        self,
        _fget: AnyFn[Concatenate[I, P], Awaitable[T]],
        field_name: Optional[str] = None,
        **modifiers: Unpack[ModifierKwargs],
    ) -> None:
        """
        Initialize the HiddenMethodDescriptor.

        Args:
            _fget: The function to be wrapped.
            field_name: Optional name for the field. If not provided, the function's name will be used.
            **modifiers: Additional modifier arguments.

        Raises:
            ValueError: If _fget is not callable.
        """
        ASyncDescriptor.__init__(self, _fget, field_name, **modifiers)
        if self.__doc__ is None:
            self.__doc__ = f"A :class:`HiddenMethodDescriptor` for :meth:`{self.__wrapped__.__qualname__}`."
        elif not self.__doc__:
            self.__doc__ += f"A :class:`HiddenMethodDescriptor` for :meth:`{self.__wrapped__.__qualname__}`."
        if self.__wrapped__.__doc__:
            self.__doc__ += f"\n\nThe original docstring for :meth:`~{self.__wrapped__.__qualname__}` is shown below:\n\n{self.__wrapped__.__doc__}"

    def __get__(_ModifiedMixin self, instance: I, owner: Type[I]) -> HiddenMethod[I, T]:
        """Retrieves the hidden method for the property.

        Args:
            instance: The instance from which the method is accessed.
            owner: The owner class of the method.

        Returns:
            The hidden method.
        """
        if instance is None:
            return self
    
        cdef object bound
        cdef str field_name = self.field_name
        try:
            bound = instance.__dict__[field_name]
        except KeyError:
            bound = HiddenMethod(
                instance,
                self.__wrapped__,
                self.__is_async_def__,
                field_name,
                **self.modifiers._modifiers,
            )
            instance.__dict__[field_name] = bound
            _logger_debug("new hidden method: %s", bound)
        _update_cache_timer(field_name, instance, bound)
        return bound


cdef object _parse_args(
    func: Union[None, DefaultMode, AsyncGetterFunction[I, T]], 
    dict modifiers,
):
    """Parses the arguments for the property decorators.

    Args:
        func: The function to be wrapped.
        modifiers: Additional modifier arguments.

    Returns:
        Tuple[Optional[AsyncGetterFunction[I, T]], ModifierKwargs] A tuple containing the parsed function and modifiers.
    """
    if func in ("sync", "async"):
        modifiers["default"] = func
        return None, modifiers
    return func, modifiers


cdef inline void _import_TaskMapping():
    global TaskMapping
    from a_sync import TaskMapping



async def sleep0() -> None:
    """Skip one event loop run cycle.

    This is a hacky helper for 'asyncio.sleep()', used
    when the 'delay' is set to 0. It yields exactly one
    time (which Task.__step knows how to handle) instead
    of creating a Future object. We monkey patch asyncio
    with our helper so it "just works" without you doing
    anything.

    While asyncio.sleep(0) is already very well-optimized,
    this equivalent helper consumes about 10% less compute
    than asyncio's equivalent internal helper.
    """



import asyncio
from asyncio import futures, tasks
from typing import Awaitable, Iterable, List, TypeVar

from a_sync import _smart
from a_sync.a_sync._helpers cimport get_event_loop


__T = TypeVar("__T")


cdef object current_task = asyncio.current_task
cdef object ensure_future = asyncio.ensure_future
cdef object CancelledError = asyncio.CancelledError
cdef object _get_loop = futures._get_loop
cdef object _GatheringFuture = tasks._GatheringFuture

cdef object smart_task_factory = _smart.smart_task_factory


def igather(
    coros_or_futures: Iterable[Awaitable[__T]], bint return_exceptions = False
) -> Awaitable[List[__T]]:
    """A clone of asyncio.gather that takes a single iterator of coroutines instead of an unpacked tuple."""
    return cigather(coros_or_futures, return_exceptions=return_exceptions)


cdef object cigather(object coros_or_futures, bint return_exceptions = False):
    """A clone of asyncio.gather that takes a single iterator of coroutines instead of an unpacked tuple."""
    # NOTE: closures inside cpdef functions not yet supported, so we have this cdef helper
    cdef long long nfuts, nfinished
    cdef dict arg_to_fut = {}
    cdef list children = []
    loop = None
    for arg in coros_or_futures:
        if arg not in arg_to_fut:
            fut = ensure_future(arg, loop=loop)
            if loop is None:
                loop = _get_loop(fut)
            if fut is not arg:
                # 'arg' was not a Future, therefore, 'fut' is a new
                # Future created specifically for 'arg'.  Since the caller
                # can't control it, disable the "destroy pending task"
                # warning.
                fut._log_destroy_pending = False

            arg_to_fut[arg] = fut

        else:
            # There's a duplicate Future object in coros_or_futures.
            fut = arg_to_fut[arg]

        children.append(fut)

    if not children:
        return _get_empty_result_set_fut(get_event_loop())

    nfuts = len(arg_to_fut)
    nfinished = 0
    outer = None  # bpo-46672

    if return_exceptions:

        def _done_callback(fut: asyncio.Future) -> None:
            # for some reason this wouldn't work until I added `return_exceptions=return_exceptions` to the func def
            nonlocal nfinished
            nfinished += 1

            if outer is None or outer.done():
                if not fut.cancelled():
                    # Mark exception retrieved.
                    fut.exception()
                return

            if nfinished == nfuts:
                # All futures are done; create a list of results
                # and set it to the 'outer' future.

                if outer._cancel_requested:
                    # If gather is being cancelled we must propagate the
                    # cancellation regardless of *return_exceptions* argument.
                    # See issue 32684.
                    try:
                        exc = fut._make_cancelled_error()
                    except AttributeError:
                        exc = CancelledError()
                    outer.set_exception(exc)
                else:
                    outer.set_result([_get_result_or_exc(child) for child in children])
    
    else:

        def _done_callback(fut: asyncio.Future) -> None:
            # for some reason this wouldn't work until I added `return_exceptions=return_exceptions` to the func def
            nonlocal nfinished
            nfinished += 1

            if outer is None or outer.done():
                if not fut.cancelled():
                    # Mark exception retrieved.
                    fut.exception()
                return

            if fut.cancelled():
                # Check if 'fut' is cancelled first, as
                # 'fut.exception()' will *raise* a CancelledError
                # instead of returning it.
                try:
                    exc = fut._make_cancelled_error()
                except AttributeError:
                    exc = CancelledError()
                outer.set_exception(exc)
                return
            else:
                exc = fut.exception()
                if exc is not None:
                    outer.set_exception(exc)
                    return

            if nfinished == nfuts:
                # All futures are done; create a list of results
                # and set it to the 'outer' future.

                if outer._cancel_requested:
                    # If gather is being cancelled we must propagate the
                    # cancellation regardless of *return_exceptions* argument.
                    # See issue 32684.
                    try:
                        exc = fut._make_cancelled_error()
                    except AttributeError:
                        exc = CancelledError()
                    outer.set_exception(exc)
                else:
                    outer.set_result([_get_result_or_exc(child) for child in children])
    
    if loop._task_factory is smart_task_factory:
        current = current_task()
        for fut in arg_to_fut.values():
            fut.add_done_callback(_done_callback)
            waiters = getattr(fut, "_waiters", None)
            if waiters is not None:
                waiters.add(current)
    else:
        for fut in arg_to_fut.values():
            fut.add_done_callback(_done_callback)

    outer = _GatheringFuture(children, loop=loop)
    
    return outer


cdef dict[object, object] _no_results_futs = {}


cdef object _get_empty_result_set_fut(loop):
    fut = _no_results_futs.get(loop)
    if fut is None:
        fut = _no_results_futs[loop] = loop.create_future()
        fut.set_result([])
    return fut


cdef object _get_result_or_exc(fut: asyncio.Future):
    if fut.cancelled():
        # Check if 'fut' is cancelled first, as 'fut.exception()'
        # will *raise* a CancelledError instead of returning it.
        # Also, since we're adding the exception return value
        # to 'results' instead of raising it, don't bother
        # setting __context__.  This also lets us preserve
        # calling '_make_cancelled_error()' at most once.
        return CancelledError("" if fut._cancel_message is None else fut._cancel_message)
    res = fut.exception()
    if res is None:
        return fut.result()
    return res



from a_sync.primitives.locks.semaphore cimport Semaphore

cdef class _AbstractPrioritySemaphore(Semaphore):
    cdef readonly dict[object, _AbstractPrioritySemaphoreContextManager] _context_managers
    cdef readonly Py_ssize_t _capacity
    cdef readonly set _potential_lost_waiters
    cdef readonly object _top_priority
    cdef readonly _AbstractPrioritySemaphoreContextManager _top_priority_manager
    cdef readonly object _context_manager_class
    cpdef object acquire(self)
    cdef _AbstractPrioritySemaphoreContextManager c_getitem(self, object priority)
    cdef dict[object, int] _count_waiters(self)
    cpdef void _wake_up_next(self)

cdef class PrioritySemaphore(_AbstractPrioritySemaphore):
    pass

cdef class _AbstractPrioritySemaphoreContextManager(Semaphore):
    cdef readonly _AbstractPrioritySemaphore _parent
    cdef readonly object _priority
    cdef readonly str _priority_name
    cpdef str _repr_no_parent_(self)

cdef class _PrioritySemaphoreContextManager(_AbstractPrioritySemaphoreContextManager):
    pass
